import itertools
from typing import Dict, Tuple, Iterable, Iterator

import numpy as np
import xarray as xr

from xcube.core.update import update_dataset_chunk_encoding


def chunk_dataset(dataset: xr.Dataset,
                  chunk_sizes: Dict[str, int] = None,
                  format_name: str = None) -> xr.Dataset:
    """
    Chunk *dataset* using *chunk_sizes* and optionally
    update encodings for given *format_name*.

    :param dataset: input dataset
    :param chunk_sizes: mapping from dimension name to new chunk size
    :param format_name: optional format, e.g. "zarr" or "netcdf4"
    :return: the (re)chunked dataset
    """
    dataset = dataset.chunk(chunks=chunk_sizes)
    if format_name:
        dataset = update_dataset_chunk_encoding(dataset,
                                                chunk_sizes=chunk_sizes,
                                                format_name=format_name)
    return dataset


def get_empty_dataset_chunks(dataset: xr.Dataset) \
        -> Iterator[Tuple[str, Iterator[Tuple[int, ...]]]]:
    """
    Identify empty dataset chunks and return their indices.

    :param dataset: The dataset.
    :return: An iterator that provides a stream of
        (variable name, block indices tuple) tuples.
    """
    return ((str(var_name), get_empty_var_chunks(dataset[var_name]))
            for var_name in dataset.data_vars)


def get_empty_var_chunks(var: xr.DataArray) -> Iterator[Tuple[int, ...]]:
    """
    Identify empty variable chunks and return their indices.

    :param var: The variable.
    :return: A list of block indices.
    """
    chunks = var.chunks
    if chunks is None:
        return None

    for chunk_index, chunk_slice in compute_chunk_slices(chunks):
        data_index = tuple(slice(start, end) for start, end in chunk_slice)
        data = var[data_index]
        if np.all(np.isnan(data)):
            # print(f'empty: {var.name}/{".".join(map(str, chunk_index))}')
            yield chunk_index


def compute_chunk_slices(chunks: Tuple[Tuple[int, ...], ...]) -> Iterable:
    chunk_indices = []
    for c in chunks:
        chunk_indices.append(tuple(i for i in range(len(c))))

    chunk_slices = []
    for c in chunks:
        x = []
        o = 0
        for s in c:
            x.append((o, o + s))
            o += s
        chunk_slices.append(tuple(x))

    return zip(itertools.product(*chunk_indices),
               itertools.product(*chunk_slices))



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import importlib
from typing import List, Any, Dict, Callable, Mapping, Sequence, Optional

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

from xcube.util.ipython import register_json_formatter

Component = Any
ComponentLoader = Callable[['Extension'], Component]
ComponentTransform = Callable[[Component, 'Extension'], Component]
ExtensionPredicate = Callable[['Extension'], bool]


class Extension:
    """
    An extension that provides a component of any type.

    Extensions are registered in a :class:`ExtensionRegistry`.

    Extension objects are not meant to be instantiated directly. Instead,
    :meth:`ExtensionRegistry.add_extension` is used to register extensions.

    :param point: extension point identifier
    :param name: extension name
    :param component: extension component
    :param loader: extension component loader function
    :param metadata: extension metadata
    """

    # noinspection PyShadowingBuiltins
    def __init__(self,
                 point: str,
                 name: str,
                 component: Component = None,
                 loader: ComponentLoader = None,
                 **metadata):

        if point is None:
            raise ValueError(f'point must be given')
        if name is None:
            raise ValueError(f'name must be given')
        if (loader is not None and component is not None) or (loader is None and component is None):
            raise ValueError(f'either component or loader must be given')
        if loader is not None and not callable(loader):
            raise ValueError(f'loader must be callable')

        self._component = component
        self._loader = loader
        self._point = point
        self._name = name
        self._metadata = metadata
        self._deleted = False

    @property
    def is_lazy(self) -> bool:
        """Whether this is a lazy extension that uses a loader."""
        return self._loader is not None

    @property
    def component(self) -> Component:
        """Extension component."""
        if self._component is None and self._loader is not None:
            self._component = self._loader(self)
        return self._component

    @property
    def point(self) -> str:
        """Extension point identifier."""
        return self._point

    @property
    def name(self) -> str:
        """Extension name."""
        return self._name

    @property
    def metadata(self) -> Dict[str, Any]:
        """Extension metadata."""
        return dict(self._metadata)

    def to_dict(self) -> Dict[str, Any]:
        """ Get a JSON-serializable dictionary representation of this extension. """

        # Note: we avoid loading the component!
        if self._component is not None:
            if hasattr(self._component, 'to_dict') and callable(getattr(self._component, 'to_dict')):
                component = self._component.to_dict()
            else:
                component = repr(self._component)
        else:
            component = '<not loaded yet>'
        d = dict(
            name=self.name,
            **self.metadata,
            point=self.point,
            component=component,
        )
        return d


register_json_formatter(Extension)


# noinspection PyShadowingBuiltins
class ExtensionRegistry:
    """
    A registry of extensions.
    Typically used by plugins to register extensions.
    """

    def __init__(self):
        self._extension_points = {}

    def has_extension(self, point: str, name: str) -> bool:
        """
        Test if an extension with given *point* and *name* is registered.

        :param point: extension point identifier
        :param name: extension name
        :return: True, if extension exists
        """
        return point in self._extension_points and name in self._extension_points[point]

    def get_extension(self, point: str, name: str) -> Optional[Extension]:
        """
        Get registered extension for given *point* and *name*.

        :param point: extension point identifier
        :param name: extension name
        :return: the extension or None, if no such exists
        """
        if point not in self._extension_points:
            return None
        return self._extension_points[point].get(name)

    def get_component(self, point: str, name: str) -> Any:
        """
        Get extension component for given *point* and *name*.
        Raises a ValueError if no such extension exists.

        :param point: extension point identifier
        :param name: extension name
        :return: extension component
        """
        extension = self.get_extension(point, name)
        if extension is None:
            raise ValueError(f'extension {name!r} not found for extension point {point!r}')
        return extension.component

    def find_extensions(self,
                        point: str,
                        predicate: ExtensionPredicate = None) -> List[Extension]:
        """
        Find extensions for *point* and optional filter function *predicate*.

        The filter function is called with an extension and should return
        a truth value to indicate a match or mismatch.

        :param point: extension point identifier
        :param predicate: optional filter function
        :return: list of matching extensions
        """
        if point not in self._extension_points:
            return []
        point_extensions = self._extension_points[point]
        if predicate is None:
            return list(point_extensions.values())
        return [extension for extension in point_extensions.values() if predicate(extension)]

    def find_components(self,
                        point: str,
                        predicate: ExtensionPredicate = None) -> List[Component]:
        """
        Find extension components for *point* and optional filter function *predicate*.

        The filter function is called with an extension and should return
        a truth value to indicate a match or mismatch.

        :param point: extension point identifier
        :param predicate: optional filter function
        :return: list of matching extension components
        """
        return [extension.component for extension in self.find_extensions(point, predicate=predicate)]

    def add_extension(self,
                      point: str,
                      name: str,
                      component: Component = None,
                      loader: ComponentLoader = None,
                      **metadata) -> Extension:
        """
        Register an extension *component* or an extension component *loader* for
        the given extension *point*, *name*, and additional *metadata*.

        Either *component* or *loader* must be specified, but not both.

        A given *loader* must be a callable with one positional argument *extension* of type :class:`Extension`
        and is expected to return the actual extension component, which may be of any type.
        The *loader* will only be called once and only when the actual extension component
        is requested for the first time. Consider using the function :func:`import_component` to create a
        loader that lazily imports a component from a module and optionally executes it.

        :param point: extension point identifier
        :param name: extension name
        :param component: extension component
        :param loader: extension component loader function
        :param metadata: extension metadata
        :return: a registered extension
        """
        extension = Extension(point, name, component=component, loader=loader, **metadata)
        if point in self._extension_points:
            self._extension_points[point][name] = extension
        else:
            self._extension_points[point] = {name: extension}
        return extension

    def remove_extension(self, point: str, name: str):
        """
        Remove registered extension *name* from given *point*.

        :param point: extension point identifier
        :param name: extension name
        """
        point_extensions = self._extension_points[point]
        del point_extensions[name]

    def to_dict(self):
        """ Get a JSON-serializable dictionary representation of this extension registry. """
        return {k: {ek: ev.to_dict() for ek, ev in v.items()} for k, v in self._extension_points.items()}


register_json_formatter(ExtensionRegistry)

_EXTENSION_REGISTRY_SINGLETON = ExtensionRegistry()


def get_extension_registry() -> ExtensionRegistry:
    """Return the extension registry singleton."""
    return _EXTENSION_REGISTRY_SINGLETON


def import_component(spec: str,
                     transform: ComponentTransform = None,
                     call: bool = False,
                     call_args: Sequence[Any] = None,
                     call_kwargs: Mapping[str, Any] = None) -> ComponentLoader:
    """
    Return a component loader that imports a module or module component from *spec*.
    To import a module, *spec* should be the fully qualified module name. To import a
    component, *spec* must also append the component name to the fully qualified module name
    separated by a color (":") character.

    An optional *transform* callable my be used to transform the imported component. If given,
    a new component is computed::

        component = transform(component, extension)

    If the *call* flag is set, the component is expected to be a callable which will be called
    using the given *call_args* and *call_kwargs* to produce a new component::

        component = component(*call_kwargs, **call_kwargs)

    Finally, the component is returned.

    :param spec: String of the form "module_path" or "module_path:component_name"
    :param transform: callable that takes two positional arguments,
        the imported component and the extension of type :class:`Extension`
    :param call: Whether to finally call the component with given *call_args* and *call_kwargs*
    :param call_args: arguments passed to a callable component if *call* flag is set
    :param call_kwargs: keyword arguments passed to callable component if *call* flag is set
    :return: a component loader
    """

    # noinspection PyUnusedLocal
    def _load(extension: Extension):
        nonlocal spec
        component = _import_component(spec, force_component=call)
        if transform is not None:
            component = transform(component, extension)
        if call or call_args or call_kwargs:
            component = component(*(call_args or []), **(call_kwargs or {}))
        return component

    return _load


def _import_component(component_spec: str, force_component: bool = False):
    """
    Import a module or module component from *spec*.

    :param component_spec: String of the form "module_name" or "module_name:component_name" where
        module_name must an absolute, fully qualified path to a module.
    :param force_component: If True, *spec* must specify a component name
    :return: the imported module or module component
    """
    if ':' in component_spec:
        module_name, component_name = component_spec.split(':', maxsplit=1)
    else:
        module_name, component_name = component_spec, None

    if force_component and component_name is None:
        raise ValueError('illegal spec, must specify a component')
    if module_name == '':
        raise ValueError('illegal spec, missing module path')
    if component_name == '':
        raise ValueError('illegal spec, missing component name')

    module = importlib.import_module(module_name)
    return getattr(module, component_name) if component_name else module



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import itertools
import json
from collections.abc import MutableMapping
from typing import Dict, Tuple, Callable, Any, Sequence
from typing import Union

import numpy as np

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

from deprecated import deprecated
from xcube.constants import LOG_LEVEL_TRACE
from xcube.core.zarrstore import LoggingZarrStore
import collections.abc
from logging import Logger
from typing import Iterator, Iterable, KeysView, Optional

from xcube.constants import LOG
from xcube.util.assertions import assert_instance

GetChunk = Callable[['ChunkStore', str, Tuple[int, ...]], bytes]


@deprecated(reason="This class shall no longer be used."
                   " If similar functionality is needed,"
                   " use xcube.core.zarrstore.GenericZarrStore",
            version="0.12.1")
class ChunkStore(MutableMapping):
    """
    A Zarr Store that generates datasets by allowing data variables to
    fetch or compute their chunks by a user-defined function *get_chunk*.
    Implements the standard Python ``MutableMapping`` interface.

    This is how the *get_chunk* function is called:::

        data = get_chunk(chunk_store, var_name, chunk_indexes)

    where ``chunk_store`` is this store, ``var_name`` is the name of
    the variable for which data is fetched, and ``chunk_indexes`` is
    a tuple of zero-based, integer chunk indexes. The result must be
    a Python *bytes* object.

    :param dims: Dimension names of all data variables,
        e.g. ('time', 'lat', 'lon').
    :param shape: Shape of all data variables according to *dims*,
        e.g. (512, 720, 1480).
    :param chunks: Chunk sizes of all data variables according to *dims*,
        e.g. (128, 180, 180).
    :param attrs: Global dataset attributes.
    :param get_chunk: Default chunk fetching/computing function.
    :param trace_store_calls: Whether to log calls
        into the ``MutableMapping`` interface.
    """

    def __init__(self,
                 dims: Sequence[str],
                 shape: Sequence[int],
                 chunks: Sequence[int],
                 attrs: Dict[str, Any] = None,
                 get_chunk: GetChunk = None,
                 trace_store_calls: bool = False):

        self._ndim = len(dims)
        self._dims = tuple(dims)
        self._shape = tuple(shape)
        self._chunks = tuple(chunks)
        self._get_chunk = get_chunk
        self._trace_store_calls = trace_store_calls

        # setup Virtual File System (vfs)
        self._vfs = {
            '.zgroup': _dict_to_bytes(dict(zarr_format=2)),
            '.zattrs': _dict_to_bytes(attrs or dict())
        }

    @property
    def ndim(self) -> int:
        return self._ndim

    @property
    def dims(self) -> Tuple[str, ...]:
        return self._dims

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape

    @property
    def chunks(self) -> Tuple[int, ...]:
        return self._chunks

    def add_array(self, name: str, array: np.ndarray, attrs: Dict):
        shape = list(map(int, array.shape))
        dtype = str(array.dtype.str)
        array_metadata = {
            "zarr_format": 2,
            "chunks": shape,
            "shape": shape,
            "dtype": dtype,
            "fill_value": None,
            "compressor": None,
            "filters": None,
            "order": "C",
        }
        self._vfs[name] = _str_to_bytes('')
        self._vfs[name + '/.zarray'] = _dict_to_bytes(array_metadata)
        self._vfs[name + '/.zattrs'] = _dict_to_bytes(attrs)
        self._vfs[name + '/' + ('.'.join(['0'] * array.ndim))] = bytes(array)

    def add_lazy_array(self,
                       name: str,
                       dtype: str,
                       fill_value: Union[int, float] = None,
                       compressor: Dict[str, Any] = None,
                       filters=None,
                       order: str = 'C',
                       attrs: Dict[str, Any] = None,
                       get_chunk: GetChunk = None):

        get_chunk = get_chunk or self._get_chunk
        if get_chunk is None:
            raise ValueError('get_chunk must be given as there is no default')

        array_metadata = dict(zarr_format=2,
                              shape=self._shape,
                              chunks=self._chunks,
                              compressor=compressor,
                              dtype=dtype,
                              fill_value=fill_value,
                              filters=filters,
                              order=order)

        self._vfs[name] = _str_to_bytes('')
        self._vfs[name + '/.zarray'] = _dict_to_bytes(array_metadata)
        self._vfs[name + '/.zattrs'] = _dict_to_bytes(
            dict(_ARRAY_DIMENSIONS=self._dims, **(attrs or dict()))
        )

        nums = np.array(self._shape) // np.array(self._chunks)
        indexes = itertools.product(*tuple(map(range, map(int, nums))))
        for index in indexes:
            filename = '.'.join(map(str, index))
            # noinspection PyTypeChecker
            self._vfs[name + '/' + filename] = name, index, get_chunk

    @property
    def _class_name(self):
        return self.__module__ + '.' + self.__class__.__name__

    ##########################################################################
    # Zarr Store (MutableMapping) implementation
    ##########################################################################

    def keys(self) -> KeysView[str]:
        if self._trace_store_calls:
            _trace(f'{self._class_name}.keys()')
        return self._vfs.keys()

    def listdir(self, key: str) -> Iterable[str]:
        if self._trace_store_calls:
            _trace(f'{self._class_name}.listdir(key={key!r})')
        if key == '':
            return (k for k in self._vfs.keys() if '/' not in k)
        else:
            prefix = key + '/'
            start = len(prefix)
            return (k for k in self._vfs.keys()
                    if k.startswith(prefix) and k.find('/', start) == -1)

    def getsize(self, key: str) -> int:
        if self._trace_store_calls:
            _trace(f'{self._class_name}.getsize(key={key!r})')
        return len(self._vfs[key])

    def __iter__(self) -> Iterator[str]:
        if self._trace_store_calls:
            _trace(f'{self._class_name}.__iter__()')
        return iter(self._vfs.keys())

    def __len__(self) -> int:
        if self._trace_store_calls:
            _trace(f'{self._class_name}.__len__()')
        return len(self._vfs.keys())

    def __contains__(self, key) -> bool:
        if self._trace_store_calls:
            _trace(f'{self._class_name}.__contains__(key={key!r})')
        return key in self._vfs

    def __getitem__(self, key: str) -> bytes:
        if self._trace_store_calls:
            _trace(f'{self._class_name}.__getitem__(key={key!r})')
        value = self._vfs[key]
        if isinstance(value, tuple):
            name, index, get_chunk = value
            return get_chunk(self, name, index)
        return value

    def __setitem__(self, key: str, value: bytes) -> None:
        if self._trace_store_calls:
            _trace(
                f'{self._class_name}.__setitem__(key={key!r}, value={value!r})')
        raise TypeError(f'{self._class_name} is read-only')

    def __delitem__(self, key: str) -> None:
        if self._trace_store_calls:
            _trace(f'{self._class_name}.__delitem__(key={key!r})')
        raise TypeError(f'{self._class_name} is read-only')


def _trace(msg: str):
    LOG.log(LOG_LEVEL_TRACE, msg)


def _dict_to_bytes(d: Dict):
    return _str_to_bytes(json.dumps(d, indent=2))


def _str_to_bytes(s: str):
    return bytes(s, encoding='utf-8')


@deprecated(reason="This class has been moved,"
                   " use xcube.core.zarrstore.LoggingZarrStore"
                   " instead.",
            version="0.12.1")
class LoggingStore(LoggingZarrStore):
    """
    A Zarr Store that logs all method calls on another store *other*
    including execution time.
    """

    @classmethod
    def new(cls,
            other: collections.abc.MutableMapping,
            logger: Logger = LOG,
            name: Optional[str] = None):
        assert_instance(other, collections.abc.MutableMapping)
        return cls(other, logger=logger, name=name)


@deprecated(reason="This class has been moved,"
                   " use xcube.core.zarrstore.LoggingZarrStore"
                   " instead.",
            version="0.12.1")
class MutableLoggingStore(LoggingStore):
    """
    Mutable version of :class:LoggingStore.
    """



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
from typing import Dict
from typing import List

DEFAULT_DEPENDENCY_NAMES = \
    ['affine', 'click', 'cmocean', 'dask', 'dask - image', 'distributed',
     'fiona', 'fsspec', 'gdal', 'geopandas', 'jdcal', 'jsonschema',
     'matplotlib - base', 'netcdf4', 'numba', 'numpy', 'pandas', 'pillow',
     'pyjwt', 'pyproj', 'pyyaml', 'rasterio', 'requests',
     'requests-oauthlib', 's3fs', 'scipy', 'setuptools', 'shapely',
     'tornado', 'urllib3', 'xarray', 'zarr']

_XCUBE_VERSIONS = None


def get_xcube_versions() -> Dict[str, str]:
    """
    Get a mapping from xcube package names to package versions.

    :return: A mapping of the package names to package versions
    The result computed from the evaluating the expression.
    """
    from .plugin import get_plugins

    global _XCUBE_VERSIONS
    if _XCUBE_VERSIONS is None:
        plugin_names = [f'{plugin}.version'
                        for plugin in list(get_plugins().keys())]

        _XCUBE_VERSIONS = get_versions(DEFAULT_DEPENDENCY_NAMES,
                                       plugin_names)
    return _XCUBE_VERSIONS


def get_versions(dependency_names: List[str], plugin_names: List[str]) \
        -> Dict[str, str]:
    """
    Get a mapping from package names to package versions.
    The input is divided into names of packages that are external dependencies
    and into names of xcube plugins.

    :param dependency_names: A list of names of packages of which the versions
        shall be found
    :param plugin_names: A list of names of xcube plugins of which the versions
        shall be found
    :return: A mapping of the package names to package versions
    The result computed from the evaluating the expression.
    """
    # Idea borrowed from xarray.print_versions
    import importlib

    def _maybe_add_dot_version(name: str):
        if not name.endswith('.version'):
            return f'{name}.version'
        return name

    dependencies = [(_maybe_add_dot_version(plugin_name),
                     lambda mod: mod.version)
                    for plugin_name in plugin_names]

    def _find_module_version(mod):
        if hasattr(mod, '__version__'):
            return mod.__version__
        elif hasattr(mod, 'version'):
            return mod.version
        else:
            return 'unknown'

    dependencies += [(dependency_name, _find_module_version)
                     for dependency_name in dependency_names]

    dependencies_dict = {}
    for (module_name, module_version) in dependencies:
        module_key = module_name.split('.')[0]
        # noinspection PyBroadException
        try:
            if module_name in sys.modules:
                module = sys.modules[module_name]
            else:
                module = importlib.import_module(module_name)
        except BaseException:
            pass
        else:
            # noinspection PyBroadException
            try:
                dependencies_dict[module_key.split('.version')[0]] \
                    = module_version(module)
            except BaseException as e:
                dependencies_dict[module_key.split('.version')[0]] = "installed"

    return dependencies_dict



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Any, Dict, List, Optional, Callable, Iterator
from typing import Sequence

import fiona
import fsspec
import pyproj
import shapely

from fiona.collection import Collection

from xcube.server.api import ApiError
from xcube.server.api import Context
from xcube.webapi.common.context import ResourcesContext

PlaceGroup = Dict[str, Any]
Feature = Dict[str, Any]

ALL_PLACES = "all"


class PlacesContext(ResourcesContext):
    _feature_index: int = 0

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self._additional_place_groups: Dict[str, List[PlaceGroup]] = dict()
        self._place_group_cache: Dict[str, PlaceGroup] = dict()

    def on_dispose(self):
        if self._place_group_cache:
            self._place_group_cache.clear()

    def get_cached_place_group(self,
                               place_group_id: str) -> Optional[PlaceGroup]:
        return self._place_group_cache.get(place_group_id)

    def set_cached_place_group(self,
                               place_group_id: str,
                               place_group: PlaceGroup):
        self._place_group_cache[place_group_id] = place_group

    def get_cached_place_groups(
            self,
            predicate: Optional[Callable[[str, PlaceGroup], bool]] = None
    ) -> List[PlaceGroup]:
        return [
            v
            for k, v in self._place_group_cache.items()
            if predicate is None or predicate(k, v)
        ]

    def new_feature_id(self) -> str:
        with self.rlock:
            feature_id = str(self._feature_index)
            self._feature_index += 1
            return feature_id

    def get_global_place_groups(self,
                                base_url: str,
                                load_features=False) -> List[PlaceGroup]:
        return self.load_place_groups(self.config.get("PlaceGroups", []),
                                      base_url,
                                      is_global=True,
                                      load_features=load_features)

    def get_global_place_group(self,
                               place_group_id: str,
                               base_url: str,
                               load_features: bool = False) -> PlaceGroup:
        place_group_config = self._get_place_group_config(place_group_id)
        return self._load_place_group(place_group_config,
                                      base_url,
                                      is_global=True,
                                      load_features=load_features)

    def _get_place_group_config(self, place_group_id: str) -> Dict:
        place_group_configs = self.config.get("PlaceGroups", [])
        for place_group_config in place_group_configs:
            if place_group_config['Identifier'] == place_group_id:
                return place_group_config
        raise ApiError.NotFound(
            f'Place group "{place_group_id}" not found'
        )

    def load_place_groups(self,
                          place_group_configs: List,
                          base_url: str,
                          is_global: bool = False,
                          load_features: bool = False,
                          qualifiers: List[str] = list()) -> List[PlaceGroup]:
        place_groups = []
        for place_group_config in place_group_configs:
            place_group = self._load_place_group(place_group_config,
                                                 base_url,
                                                 is_global=is_global,
                                                 load_features=load_features)
            place_groups.append(place_group)
        for q in [qualifier for qualifier in qualifiers
                  if qualifier in self._additional_place_groups]:
            for place_group in self._additional_place_groups[q]:
                place_groups.append(place_group)
        return place_groups

    def add_place_group(self,
                        place_group: PlaceGroup,
                        qualifiers: List[str] = list()):
        for qualifier in qualifiers:
            if qualifier not in self._additional_place_groups:
                self._additional_place_groups[qualifier] = []
            self._additional_place_groups[qualifier].append(place_group)

    def _load_place_group(self,
                          place_group_config: Dict[str, Any],
                          base_url: str,
                          is_global: bool = False,
                          load_features: bool = False) -> PlaceGroup:
        place_group_id = place_group_config.get("PlaceGroupRef")
        if place_group_id:
            if is_global:
                raise ApiError.InvalidServerConfig(
                    "'PlaceGroupRef' cannot be used in a global place group"
                )
            if len(place_group_config) > 1:
                raise ApiError.InvalidServerConfig(
                    "'PlaceGroupRef' if present,"
                    " must be the only entry in a 'PlaceGroups' item"
                )
            return self.get_global_place_group(place_group_id, base_url,
                                               load_features=load_features)

        place_group_id = self.get_place_group_id_safe(place_group_config)

        place_group = self.get_cached_place_group(place_group_id)
        if place_group is None:
            place_group_title = place_group_config.get("Title",
                                                       place_group_id)
            place_path_wc = self.get_config_path(place_group_config,
                                                 f"'PlaceGroups' item")
            fs, place_path = fsspec.core.url_to_fs(place_path_wc)
            source_paths = [fs.unstrip_protocol(p)
                            for p in fs.glob(place_path)]
            source_encoding = place_group_config.get("CharacterEncoding",
                                                     "utf-8")

            join = None
            place_join = place_group_config.get("Join")
            if isinstance(place_join, dict):
                join_path = self.get_config_path(
                    place_join, "'Join' of a 'PlaceGroups' item"
                )
                join_property = place_join.get("Property")
                if not join_property:
                    raise ApiError.InvalidServerConfig(
                        "Missing 'Property' entry in 'Join'"
                        " of a 'PlaceGroups' item"
                    )
                join_encoding = place_join.get("CharacterEncoding", "utf-8")
                join = dict(path=join_path, property=join_property,
                            encoding=join_encoding)

            property_mapping = self.get_property_mapping(base_url,
                                                         place_group_config)

            place_group = dict(type="FeatureCollection",
                               features=None,
                               id=place_group_id,
                               title=place_group_title,
                               propertyMapping=property_mapping,
                               sourcePaths=source_paths,
                               sourceEncoding=source_encoding,
                               join=join)

            self.check_sub_group_configs(place_group_config)
            self.set_cached_place_group(place_group_id, place_group)

        if load_features:
            self.load_place_group_features(place_group)

        return place_group

    @staticmethod
    def get_place_group_id_safe(place_group_config):
        place_group_id = place_group_config.get("Identifier")
        if not place_group_id:
            raise ApiError.InvalidServerConfig(
                "Missing 'Identifier' entry in a 'PlaceGroups' item"
            )
        return place_group_id

    @staticmethod
    def check_sub_group_configs(place_group_config):
        sub_place_group_configs = place_group_config.get("Places")
        if sub_place_group_configs:
            raise ApiError.InvalidServerConfig(
                "Invalid 'Places' entry in a 'PlaceGroups' item:"
                " not implemented yet"
            )

    @staticmethod
    def get_property_mapping(base_url, place_group_config):
        property_mapping = place_group_config.get("PropertyMapping")
        if property_mapping:
            property_mapping = dict(property_mapping)
            for key, value in property_mapping.items():
                if isinstance(value, str) and '${base_url}' in value:
                    property_mapping[key] = value.replace('${base_url}',
                                                          base_url)
        return property_mapping

    def load_place_group_features(self, place_group: PlaceGroup) \
            -> List[Dict[str, Any]]:
        features = place_group.get('features')
        if features is not None:
            return features
        source_paths = place_group['sourcePaths']
        source_encoding = place_group['sourceEncoding']
        features = []
        for source_path in source_paths:
            with self.measure_time(
                    f'Loading features from file {source_path}'):
                with fiona.open(source_path,
                                encoding=source_encoding) \
                        as feature_collection:
                    for feature in self._to_geo_interface(feature_collection):
                        self._remove_feature_id(feature)
                        feature["id"] = self.new_feature_id()
                        features.append(feature)

        join = place_group['join']
        if join:
            join_path = join['path']
            join_property = join['property']
            join_encoding = join['encoding']
            with fiona.open(join_path,
                            encoding=join_encoding) as feature_collection:
                indexed_join_features = self._get_indexed_features(
                    list(self._to_geo_interface(feature_collection)),
                    join_property
                )
            for feature in features:
                properties = feature.get('properties')
                if isinstance(properties, dict) \
                        and join_property in properties:
                    join_value = properties[join_property]
                    join_feature = indexed_join_features.get(join_value)
                    if join_feature:
                        join_properties = join_feature.get('properties')
                        if join_properties:
                            properties.update(join_properties)
                            feature['properties'] = properties

        place_group['features'] = features
        return features

    @classmethod
    def _to_geo_interface(cls, feature_collection: Collection) \
            -> Iterator[Dict[str, Any]]:
        source_crs = feature_collection.crs
        target_crs = fiona.crs.CRS.from_epsg(4326)
        if not source_crs == target_crs:
            project = pyproj.Transformer.from_crs(
                source_crs, target_crs, always_xy=True).transform
        for feature in feature_collection:
            # fiona >=1.9 returns features of type fiona.model.Feature
            # rather than JSON-serializable dictionaries.
            if hasattr(feature, '__geo_interface__'):
                # We fall back on the traditional geo-interface:
                feature = feature.__geo_interface__
                # Fiona =1.9.0 adds empty "geometries" field
                # to any "geometry", we fix this too:
                geometry = feature.get('geometry')
                if geometry \
                        and "geometries" in geometry \
                        and geometry.get("type") != "GeometryCollection":
                    del geometry["geometries"]
            if not source_crs == target_crs:
                geometry = feature.get('geometry')
                shapely_geom = shapely.geometry.shape(geometry)
                feature['geometry'] = (shapely.ops.transform(
                    project, shapely_geom).__geo_interface__)
            yield feature

    @classmethod
    def _get_indexed_features(cls,
                              features: Sequence[Feature],
                              property_name: str) -> Feature:
        feature_index = {}
        for feature in features:
            properties = feature.get('properties')
            if properties and property_name in properties:
                property_value = properties[property_name]
                feature_index[property_value] = feature
        return feature_index

    @classmethod
    def _remove_feature_id(cls, feature: Feature):
        if "id" in feature:
            del feature["id"]
        if "ID" in feature:
            del feature["ID"]



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from .cached import CachedZarrStore
from .diagnostic import DiagnosticZarrStore
from .generic import GenericArray
from .generic import GenericArrayLike
from .generic import GenericZarrStore
from .holder import ZarrStoreHolder
from .logging import LoggingZarrStore



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
import warnings
from typing import Optional, Union, Dict, Tuple, Sequence, Any, Mapping, List

import affine
import dask.array as da
import geopandas as gpd
import numpy as np
import rasterio.features
import shapely.geometry
import shapely.geometry
import shapely.wkt
import xarray as xr
from deprecated import deprecated

from xcube.core.schema import get_dataset_bounds_var_name
from xcube.core.schema import get_dataset_chunks
from xcube.core.schema import get_dataset_xy_var_names
from xcube.core.update import update_dataset_spatial_attrs
from xcube.util.geojson import GeoJSON
from xcube.util.types import normalize_scalar_or_pair

GeometryLike = Union[shapely.geometry.base.BaseGeometry,
                     Dict[str, Any],
                     str,
                     Sequence[Union[float, int]]]
Bounds = Tuple[float, float, float, float]
SplitBounds = Tuple[Bounds, Optional[Bounds]]

Name = str
Attrs = Mapping[Name, Any]
GeoJSONFeature = Mapping[Name, Any]
GeoJSONFeatures = Sequence[GeoJSONFeature]
GeoDataFrame = 'pandas.geodataframe.GeoDataFrame'
VarProps = Mapping[Name, Mapping[Name, Any]]

_INVALID_GEOMETRY_MSG = (
    'Geometry must be either a shapely geometry object, '
    'a GeoJSON-serializable dictionary, a geometry WKT string, '
    'box coordinates (x1, y1, x2, y2), '
    'or point coordinates (x, y)'
)

_INVALID_BOX_COORDS_MSG = 'Invalid box coordinates'


def rasterize_features(
        dataset: xr.Dataset,
        features: Union[GeoDataFrame, GeoJSONFeatures],
        feature_props: Sequence[Name],
        var_props: Dict[Name, VarProps] = None,
        tile_size: Union[int, Tuple[int, int]] = None,
        all_touched: bool = False,
        in_place: bool = False,
) -> Optional[xr.Dataset]:
    """
    Rasterize feature properties given by *feature_props* of
    vector-data *features* as new variables into *dataset*.

    *dataset* must have two spatial 1-D coordinates, either
    ``lon`` and ``lat`` in degrees, reprojected coordinates,
    ``x`` and ``y``, or similar.

    *feature_props* is a sequence of names of feature properties
    that must exists in each feature of *features*.

    *features* may be passed as pandas.GeoDataFrame`` or as an
    iterable of GeoJSON features.

    Using the optional *var_props*, the properties of newly
    created variables from feature properties can be specified.
    It is a mapping of feature property names to mappings of
    variable properties.
    Here is an example variable properties mapping:::

    {
        'name': 'land_class',  # (str) - the variable's name,
                               # default is the feature property name;
        'dtype' np.int16,      # (str|np.dtype) - the variable's dtype,
                               # default is np.float64;
        'fill_value': -999,    # (bool|int|float|np.nparray) -
                               # the variable's fill value,
                               # default is np.nan;
        'attrs': {},           # (Mapping[str, Any]) -
                               # the variable's fill value, default is {};
        'converter': int,      # (Callable[[Any], Any]) -
                               # a converter function used to convert
                               # from property feature value to variable
                               # value, default is float.
                               # Deprecated, no longer used.
    }

    Note that newly created variables will have data type `np.float64`
    because `np.nan` is used to encode missing values. `fill_value` and
    `dtype` are used to encode the variables when persisting the data.

    Currently, the coordinates of the geometries in the given
    *features* must use the same CRS as the given *dataset*.

    :param dataset: The xarray dataset.
    :param features: A ``geopandas.GeoDataFrame`` instance
        or a sequence of GeoJSON features.
    :param feature_props: Sequence of names of numeric feature
        properties to be rasterized.
    :param var_props: Optional mapping of feature property name
        to a name or a 5-tuple (name, dtype, fill_value,
        attributes, converter) for the new variable.
    :param tile_size: If given, the unconditional spatial chunk sizes
        in x- and y-direction in pixels.
        May be given as integer scalar or x,y-pair of integers.
    :param all_touched: If True, all pixels intersected by a
        feature's geometry outlines will be included.
        If False, only pixels whose center is
        within the feature polygon or that are selected by Bresenham’s line
        algorithm will be included in the mask.
        The default is False.
    :param in_place: Whether to add new variables to *dataset*.
        If False, a copy will be created and returned.
    :return: dataset with rasterized feature_property
    """

    var_props = var_props or {}
    for v in var_props.values():
        if v and 'converter' in v:
            warnings.warn(f'the "converter" property of var_props'
                          f' has been deprecated and will be ignored',
                          DeprecationWarning)

    xy_var_names = get_dataset_xy_var_names(dataset, must_exist=True)
    x_min, y_min, x_max, y_max = get_dataset_bounds(dataset)
    x_var_name, y_var_name = xy_var_names
    x_var, y_var = dataset[x_var_name], dataset[y_var_name]
    x_dim, y_dim = x_var.dims[0], y_var.dims[0]
    yx_coords = {y_var_name: y_var, x_var_name: x_var}
    yx_dims = y_dim, x_dim

    width = x_var.size
    height = y_var.size
    x_res = (x_max - x_min) / width
    y_res = (y_max - y_min) / height

    yx_chunks = _get_spatial_chunks(dataset,
                                    x_var_name, y_var_name,
                                    tile_size)

    if isinstance(features, gpd.GeoDataFrame):
        geo_data_frame = features
    else:
        geo_data_frame = gpd.GeoDataFrame.from_features(features)
    for feature_prop_name in feature_props:
        if feature_prop_name not in geo_data_frame:
            raise ValueError(f'feature property '
                             f'{feature_prop_name!r} not found')

    # Filter out empty or invalid geometries, remember valid rows
    geometries = []
    valid_row_indices = []
    for row_index in range(len(geo_data_frame)):
        geometry = geo_data_frame.geometry[row_index]
        if not geometry.is_empty and geometry.is_valid:
            valid_row_indices.append(row_index)
            geometries.append(geometry.__geo_interface__)

    # Collect columns of feature data for valid row indices
    valid_row_indices = np.array(valid_row_indices)
    feature_data = []
    for feature_prop_name in feature_props:
        feature_values = geo_data_frame[feature_prop_name].to_numpy()
        feature_values = feature_values[valid_row_indices]
        var_prop_mapping = var_props.get(feature_prop_name, {})
        dtype = var_prop_mapping.get('dtype', feature_values.dtype)
        if dtype != feature_values.dtype:
            feature_values = feature_values.astype(dtype=dtype)
        feature_data.append(feature_values)

    # Rasterize all features into single blocks of type np.float64
    #
    # Note, we process all rows and all features in every block,
    # so we need to mask only once per block and because masking
    # is an expensive operation.
    #
    # When there are many feature variables, and they are accessed
    # independently of each other, this may quickly become
    # expensive in terms of CPU.
    # However, if we'd create dask arrays for every row, this will become
    # also expensive in terms of memory and CPU, because of the potentially
    # very large graphs.
    # Alternatively, we could create independent dask graphs for
    # every feature with all rows processed in each block.
    #
    num_features = len(feature_props)
    chunks = da.core.normalize_chunks((num_features, *yx_chunks),
                                      (num_features, height, width))
    rasterized_features = da.map_blocks(
        _rasterize_features_into_block,
        chunks=chunks,
        dtype=np.float64,
        meta=np.array((), dtype=np.float64),
        geometries=geometries,
        feature_data=feature_data,
        x_offset=x_min,
        y_offset=y_max,
        x_res=x_res,
        y_res=y_res,
        all_touched=all_touched
    )

    if y_var[0] < y_var[-1]:
        rasterized_features = rasterized_features[:, ::-1, ::]

    if not in_place:
        dataset = xr.Dataset(coords=dataset.coords, attrs=dataset.attrs)

    # Create feature variables from rasterized features
    for feature_index, feature_prop_name in enumerate(feature_props):
        var_prop_mapping = var_props.get(feature_prop_name, {})
        var_name = var_prop_mapping.get(
            'name',
            feature_prop_name.replace(' ', '_')
        )
        var_dtype = np.dtype(var_prop_mapping.get('dtype', np.float64))
        var_fill_value = var_prop_mapping.get('fill_value', np.nan)
        var_attrs = var_prop_mapping.get('attrs', {})

        feature_image = rasterized_features[feature_index]

        feature_var = xr.DataArray(feature_image,
                                   coords=yx_coords,
                                   dims=yx_dims,
                                   attrs=var_attrs)
        feature_var.encoding.update(_FillValue=var_fill_value,
                                    dtype=var_dtype)
        dataset[var_name] = feature_var

    return dataset


def _rasterize_features_into_block(
        block_info: Dict[Union[str, None], Any] = None,
        geometries: List[Dict[str, Any]] = None,
        feature_data: List[np.ndarray] = None,
        x_offset: float = None,
        y_offset: float = None,
        x_res: float = None,
        y_res: float = None,
        all_touched: bool = None
):
    ret_info = block_info[None]
    dtype = ret_info['dtype']
    chunk_shape = ret_info['chunk-shape']
    num_features, height, width = chunk_shape
    image_shape = height, width
    _, (y_start, y_end), (x_start, x_end) = ret_info['array-location']
    x1 = x_offset + x_res * x_start
    x2 = x_offset + x_res * x_end
    y1 = y_offset - y_res * y_start
    y2 = y_offset - y_res * y_end
    transform = affine.Affine(x_res, 0.0, x1,
                              0.0, -y_res, y1)
    block_bounds = shapely.geometry.box(x1, min(y1, y2),
                                        x2, max(y1, y2))
    block = np.full(chunk_shape, np.nan, dtype=dtype)
    for row_index, geometry in enumerate(geometries):
        shape = shapely.geometry.shape(geometry)
        shape = block_bounds.intersection(shape)
        if shape.is_empty:
            continue
        if not shape.is_valid:
            continue
        mask = rasterio.features.geometry_mask([shape],
                                               out_shape=image_shape,
                                               transform=transform,
                                               all_touched=all_touched,
                                               invert=True)
        for i in range(num_features):
            background = block[i]
            foreground = np.full(image_shape,
                                 feature_data[i][row_index],
                                 dtype=dtype)
            block[i, :, :] = np.where(mask, foreground, background)

    return block


def mask_dataset_by_geometry(
        dataset: xr.Dataset,
        geometry: GeometryLike,
        tile_size: Union[int, Tuple[int, int]] = None,
        excluded_vars: Sequence[str] = None,
        no_clip: bool = False,
        all_touched: bool = False,
        save_geometry_mask: Union[str, bool] = False,
        save_geometry_wkt: Union[str, bool] = False
) -> Optional[xr.Dataset]:
    """
    Mask a dataset according to the given geometry. The cells of
    variables of the returned dataset will have NaN-values where their
    spatial coordinates are not intersecting
    the given geometry.

    :param dataset: The dataset
    :param geometry: A geometry-like object,
        see py:function:`convert_geometry`.
    :param tile_size: If given, the unconditional spatial chunk sizes
        in x- and y-direction in pixels.
        May be given as integer scalar or x,y-pair of integers.
    :param excluded_vars: Optional sequence of names of data variables
        that should not be masked (but still may be clipped).
    :param no_clip: If True, the function will not clip the dataset
        before masking, this is, the returned dataset will have the same
        dimension size as the given *dataset*.
    :param all_touched: If True, all pixels intersected by geometry outlines
        will be included in the mask. If False, only pixels whose center is
        within the polygon or that are selected by Bresenham’s line
        algorithm will be included in the mask.
        The default value is set to `False`.
    :param save_geometry_mask: If the value is a string,
        the effective geometry mask array is stored as a 2D data variable
        named by *save_geometry_mask*. If the value is True,
        the name "geometry_mask" is used.
    :param save_geometry_wkt: If the value is a string,
        the effective intersection geometry is stored as
        a Geometry WKT string in the global attribute named
        by *save_geometry*.
        If the value is True, the name "geometry_wkt" is used.
    :return: The dataset spatial subset, or None if the bounding box
        of the dataset has a no or a zero area intersection with the
        bounding box of the geometry.
    """
    geometry = normalize_geometry(geometry)
    xy_var_names = get_dataset_xy_var_names(dataset, must_exist=True)
    dataset_bounds = get_dataset_bounds(dataset, xy_var_names=xy_var_names)
    intersection_geometry = intersect_geometries(
        dataset_bounds,
        geometry
    )
    if intersection_geometry is None:
        return None

    if not no_clip:
        dataset = _clip_dataset_by_geometry(dataset,
                                            intersection_geometry,
                                            xy_var_names)

    x_min, y_min, x_max, y_max = get_dataset_bounds(
        dataset, xy_var_names=xy_var_names
    )

    x_var_name, y_var_name = xy_var_names
    x_var, y_var = dataset[x_var_name], dataset[y_var_name]

    width = x_var.size
    height = y_var.size
    x_res = (x_max - x_min) / width
    y_res = (y_max - y_min) / height

    yx_chunks = _get_spatial_chunks(dataset,
                                    x_var_name, y_var_name,
                                    tile_size)

    chunks = da.core.normalize_chunks(yx_chunks, shape=(height, width))

    mask_data = da.map_blocks(
        _mask_block,
        chunks=chunks,
        dtype=bool,
        meta=np.array((), dtype=bool),
        geometry=intersection_geometry,
        x_offset=x_min,
        y_offset=y_max,
        x_res=x_res,
        y_res=y_res,
        all_touched=all_touched
    )

    if y_var[0] < y_var[-1]:
        mask_data = mask_data[::-1, ::]

    mask = xr.DataArray(mask_data,
                        coords={y_var_name: y_var, x_var_name: x_var},
                        dims=(y_var.dims[0], x_var.dims[0]))

    dataset_vars = {}
    for var_name, var in dataset.data_vars.items():
        if not excluded_vars or var_name not in excluded_vars:
            dataset_vars[var_name] = var.where(mask)
        else:
            dataset_vars[var_name] = var

    masked_dataset = xr.Dataset(dataset_vars,
                                coords=dataset.coords,
                                attrs=dataset.attrs)

    _save_geometry_mask(masked_dataset,
                        mask,
                        save_geometry_mask)
    _save_geometry_wkt(masked_dataset,
                       intersection_geometry,
                       save_geometry_wkt)

    return masked_dataset


def _mask_block(
        block_info: Dict[Union[str, None], Any] = None,
        geometry: Dict[str, Any] = None,
        x_offset: float = None,
        y_offset: float = None,
        x_res: float = None,
        y_res: float = None,
        all_touched: bool = None
):
    ret_info = block_info[None]
    height, width = ret_info['chunk-shape']
    (y_start, _), (x_start, _) = ret_info['array-location']
    x1 = x_offset + x_res * x_start
    y1 = y_offset - y_res * y_start
    transform = affine.Affine(x_res, 0.0, x1,
                              0.0, -y_res, y1)
    return rasterio.features.geometry_mask(
        [shapely.geometry.shape(geometry)],
        out_shape=(height, width),
        transform=transform,
        all_touched=all_touched,
        invert=True
    )


def _get_spatial_chunks(dataset: xr.Dataset,
                        x_var_name: str,
                        y_var_name: str,
                        tile_size: Union[None, int, Tuple[int, int]]):
    width = dataset[x_var_name].size
    height = dataset[y_var_name].size
    if tile_size:
        tile_size = normalize_scalar_or_pair(
            tile_size,
            item_type=int,
            name='tile_size'
        )
        yx_chunks = (min(height, tile_size[1]), min(width, tile_size[0]))
    else:
        dataset_chunks = get_dataset_chunks(dataset)
        yx_chunks = (dataset_chunks.get(y_var_name),
                     dataset_chunks.get(x_var_name))
        if not all(yx_chunks):
            yx_chunks = (min(height, 1024), min(width, 1024))
    return yx_chunks


def clip_dataset_by_geometry(
        dataset: xr.Dataset,
        geometry: GeometryLike,
        save_geometry_wkt: Union[str, bool] = False
) -> Optional[xr.Dataset]:
    """
    Spatially clip a dataset according to the bounding box of a
    given geometry.

    :param dataset: The dataset
    :param geometry: A geometry-like object,
        see py:function:`convert_geometry`.
    :param save_geometry_wkt: If the value is a string,
        the effective intersection geometry is stored as
        a Geometry WKT string in the global attribute named
        by *save_geometry*. If the value is True, the name
        "geometry_wkt" is used.
    :return: The dataset spatial subset, or None if the bounding
        box of the dataset has a no or a zero area
        intersection with the bounding box of the geometry.
    """
    xy_var_names = get_dataset_xy_var_names(dataset, must_exist=True)
    intersection_geometry = intersect_geometries(
        get_dataset_bounds(dataset, xy_var_names=xy_var_names),
        geometry
    )
    if intersection_geometry is None:
        return None
    return _clip_dataset_by_geometry(
        dataset,
        intersection_geometry,
        xy_var_names,
        save_geometry_wkt=save_geometry_wkt
    )


def _clip_dataset_by_geometry(
        dataset: xr.Dataset,
        intersection_geometry: shapely.geometry.base.BaseGeometry,
        xy_var_names: Tuple[str, str],
        save_geometry_wkt: bool = False
) -> Optional[xr.Dataset]:
    # TODO (forman): the following code is wrong,
    #   if the dataset bounds cross the anti-meridian!

    ds_x_min, ds_y_min, ds_x_max, ds_y_max = get_dataset_bounds(
        dataset, xy_var_names=xy_var_names
    )

    x_var_name, y_var_name = xy_var_names
    x_var = dataset[x_var_name]
    y_var = dataset[y_var_name]

    width = x_var.size
    height = y_var.size
    res_x = (ds_x_max - ds_x_min) / width
    res_y = (ds_y_max - ds_y_min) / height

    g_x_min, g_y_min, g_x_max, g_y_max = intersection_geometry.bounds
    x1 = _clamp(int(math.floor((g_x_min - ds_x_min) / res_x)), 0, width - 1)
    x2 = _clamp(int(math.ceil((g_x_max - ds_x_min) / res_x)), 0, width - 1)
    y1 = _clamp(int(math.floor((g_y_min - ds_y_min) / res_y)), 0, height - 1)
    y2 = _clamp(int(math.ceil((g_y_max - ds_y_min) / res_y)), 0, height - 1)
    if y_var[0] > y_var[-1]:  # inverse ?
        _y1, _y2 = y1, y2
        y1 = height - _y2 - 1
        y2 = height - _y1 - 1

    dataset_subset = dataset.isel(**{x_var_name: slice(x1, x2),
                                     y_var_name: slice(y1, y2)})

    update_dataset_spatial_attrs(dataset_subset,
                                 update_existing=True,
                                 in_place=True)

    _save_geometry_wkt(dataset_subset,
                       intersection_geometry,
                       save_geometry_wkt)

    return dataset_subset


def _save_geometry_mask(dataset, mask, save_mask):
    if save_mask:
        var_name = save_mask \
            if isinstance(save_mask, str) \
            else 'geometry_mask'
        dataset[var_name] = mask


def _save_geometry_wkt(dataset, intersection_geometry, save_geometry):
    if save_geometry:
        attr_name = save_geometry \
            if isinstance(save_geometry, str) \
            else 'geometry_wkt'
        dataset.attrs.update({attr_name: intersection_geometry.wkt})


def intersect_geometries(geometry1: GeometryLike, geometry2: GeometryLike) \
        -> Optional[shapely.geometry.base.BaseGeometry]:
    geometry1 = normalize_geometry(geometry1)
    if geometry1 is None:
        return None
    geometry2 = normalize_geometry(geometry2)
    if geometry2 is None:
        return geometry1
    intersection_geometry = geometry1.intersection(geometry2)
    if not intersection_geometry.is_valid or intersection_geometry.is_empty:
        return None
    return intersection_geometry


def normalize_geometry(geometry: Optional[GeometryLike]) \
        -> Optional[shapely.geometry.base.BaseGeometry]:
    """
    Convert a geometry-like object into a shapely geometry
    object (``shapely.geometry.BaseGeometry``).

    A geometry-like object may be any shapely geometry object,
    * a dictionary that can be serialized to valid GeoJSON,
    * a WKT string,
    * a box given by a string of the form "<x1>,<y1>,<x2>,<y2>"
      or by a sequence of four numbers x1, y1, x2, y2,
    * a point by a string of the form "<x>,<y>"
      or by a sequence of two numbers x, y.

    Handling of geometries crossing the anti-meridian:

    * If box coordinates are given, it is allowed to pass
      x1, x2 where x1 > x2, which is interpreted as a box crossing
      the anti-meridian. In this case the function splits the box
      along the anti-meridian and returns a multi-polygon.
    * In all other cases, 2D geometries are assumed to _not cross
      the anti-meridian at all_.

    :param geometry: A geometry-like object
    :return: Shapely geometry object or None.
    """

    if isinstance(geometry, shapely.geometry.base.BaseGeometry):
        return geometry

    if isinstance(geometry, dict):
        if GeoJSON.is_geometry(geometry):
            return shapely.geometry.shape(geometry)
        elif GeoJSON.is_feature(geometry):
            geometry = GeoJSON.get_feature_geometry(geometry)
            if geometry is not None:
                return shapely.geometry.shape(geometry)
        elif GeoJSON.is_feature_collection(geometry):
            features = GeoJSON.get_feature_collection_features(geometry)
            if features is not None:
                geometries = [
                    f2
                    for f2 in [
                        GeoJSON.get_feature_geometry(f1)
                        for f1 in features
                    ]
                    if f2 is not None
                ]
                if geometries:
                    geometry = dict(type='GeometryCollection',
                                    geometries=geometries)
                    return shapely.geometry.shape(geometry)
        raise ValueError(_INVALID_GEOMETRY_MSG)

    if isinstance(geometry, str):
        return shapely.wkt.loads(geometry)

    if geometry is None:
        return None

    invalid_box_coords = False
    # noinspection PyBroadException
    try:
        x1, y1, x2, y2 = geometry
        is_point = x1 == x2 and y1 == y2
        if is_point:
            return shapely.geometry.Point(x1, y1)
        invalid_box_coords = x1 == x2 or y1 >= y2
        if not invalid_box_coords:
            return get_box_split_bounds_geometry(x1, y1, x2, y2)
    except Exception:
        # noinspection PyBroadException
        try:
            x, y = geometry
            return shapely.geometry.Point(x, y)
        except Exception:
            pass

    if invalid_box_coords:
        raise ValueError(_INVALID_BOX_COORDS_MSG)
    raise ValueError(_INVALID_GEOMETRY_MSG)


@deprecated(version="0.11.2",
            reason="convert_geometry() has been"
                   " renamed to normalize_geometry()")
def convert_geometry(geometry: Optional[GeometryLike]) \
        -> Optional[shapely.geometry.base.BaseGeometry]:
    return normalize_geometry(geometry)


convert_geometry.__doc__ = normalize_geometry.__doc__


@deprecated(version="0.11.2",
            reason='Uses wrong definition of "inverted". No longer used.')
def is_dataset_y_axis_inverted(
        dataset: Union[xr.Dataset, xr.DataArray],
        xy_var_names: Tuple[str, str] = None
) -> bool:
    if xy_var_names is None:
        xy_var_names = get_dataset_xy_var_names(dataset, must_exist=True)
    y_var = dataset[xy_var_names[1]]
    return float(y_var[0]) < float(y_var[-1])


def is_lon_lat_dataset(dataset: Union[xr.Dataset, xr.DataArray],
                       xy_var_names: Tuple[str, str] = None) -> bool:
    if xy_var_names is None:
        xy_var_names = get_dataset_xy_var_names(dataset, must_exist=True)
    x_var_name, y_var_name = xy_var_names
    if x_var_name == 'lon' and y_var_name == 'lat':
        return True
    x_var = dataset[x_var_name]
    y_var = dataset[y_var_name]
    return x_var.attrs.get('long_name') == 'longitude' \
           and y_var.attrs.get('long_name') == 'latitude'


def get_dataset_geometry(dataset: Union[xr.Dataset, xr.DataArray],
                         xy_var_names: Tuple[str, str] = None) \
        -> shapely.geometry.base.BaseGeometry:
    if xy_var_names is None:
        xy_var_names = get_dataset_xy_var_names(dataset, must_exist=True)
    geo_bounds = get_dataset_bounds(dataset, xy_var_names=xy_var_names)
    if is_lon_lat_dataset(dataset, xy_var_names=xy_var_names):
        return get_box_split_bounds_geometry(*geo_bounds)
    else:
        return shapely.geometry.box(*geo_bounds)


def get_dataset_bounds(dataset: Union[xr.Dataset, xr.DataArray],
                       xy_var_names: Optional[Tuple[str, str]] = None) -> Bounds:
    if xy_var_names is None:
        xy_var_names = get_dataset_xy_var_names(dataset, must_exist=True)
    x_name, y_name = xy_var_names
    x_var, y_var = dataset.coords[x_name], dataset.coords[y_name]
    is_lon = xy_var_names[0] == 'lon'

    # Note, x_min > x_max then we intersect with the anti-meridian
    x_bnds_name = get_dataset_bounds_var_name(dataset, x_name)
    if x_bnds_name is not None:
        x_bnds_var = dataset[x_bnds_name]
        x1 = x_bnds_var[0, 0]
        x2 = x_bnds_var[0, 1]
        x3 = x_bnds_var[-1, 0]
        x4 = x_bnds_var[-1, 1]
        x_min = min(x1, x2)
        x_max = max(x3, x4)
    else:
        x_min = x_var[0]
        x_max = x_var[-1]
        delta = (x_max - x_min
                 + (0 if (x_max >= x_min or not is_lon)
                    else 360)) / (x_var.size - 1)
        x_min -= 0.5 * delta
        x_max += 0.5 * delta

    # Note, x-axis may be inverted
    y_bnds_name = get_dataset_bounds_var_name(dataset, y_name)
    if y_bnds_name is not None:
        y_bnds_var = dataset[y_bnds_name]
        y1 = y_bnds_var[0, 0]
        y2 = y_bnds_var[0, 1]
        y3 = y_bnds_var[-1, 0]
        y4 = y_bnds_var[-1, 1]
        y_min = min(y1, y2, y3, y4)
        y_max = max(y1, y2, y3, y4)
    else:
        y1 = y_var[0]
        y2 = y_var[-1]
        delta = abs(y2 - y1) / (y_var.size - 1)
        y_min = min(y1, y2) - 0.5 * delta
        y_max = max(y1, y2) + 0.5 * delta

    return float(x_min), float(y_min), float(x_max), float(y_max)


def get_box_split_bounds(lon_min: float, lat_min: float,
                         lon_max: float, lat_max: float) -> SplitBounds:
    if lon_max >= lon_min:
        return ((lon_min, lat_min, lon_max, lat_max),
                None)
    else:
        return ((lon_min, lat_min, 180.0, lat_max),
                (-180.0, lat_min, lon_max, lat_max))


def get_box_split_bounds_geometry(
        lon_min: float, lat_min: float,
        lon_max: float, lat_max: float
) -> shapely.geometry.base.BaseGeometry:
    box_1, box_2 = get_box_split_bounds(lon_min, lat_min, lon_max, lat_max)
    if box_2 is not None:
        return shapely.geometry.MultiPolygon(
            polygons=[shapely.geometry.box(*box_1),
                      shapely.geometry.box(*box_2)]
        )
    else:
        return shapely.geometry.box(*box_1)


def _clamp(x, x1, x2):
    if x < x1:
        return x1
    if x > x2:
        return x2
    return x



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonComplexSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.webapi.common.schemas import BOOLEAN_SCHEMA
from xcube.webapi.common.schemas import CHUNK_SIZE_SCHEMA
from xcube.webapi.common.schemas import FILE_SYSTEM_SCHEMA
from xcube.webapi.common.schemas import GEO_BOUNDING_BOX_SCHEMA
from xcube.webapi.common.schemas import IDENTIFIER_SCHEMA
from xcube.webapi.common.schemas import PATH_SCHEMA
from xcube.webapi.common.schemas import STRING_SCHEMA
from xcube.webapi.common.schemas import URI_SCHEMA
from ..places.config import PLACE_GROUP_SCHEMA

ATTRIBUTION_SCHEMA = JsonComplexSchema(
    one_of=[
        STRING_SCHEMA,
        JsonArraySchema(items=STRING_SCHEMA),
    ]
)

VARIABLES_SCHEMA = JsonArraySchema(
    items=IDENTIFIER_SCHEMA,
    min_items=1,
    description="Names of variables to be published."
                " Names may use wildcard characters '*' and '?'."
                " Also determines the order of variables."
)

VALUE_RANGE_SCHEMA = JsonArraySchema(items=[
    JsonNumberSchema(),
    JsonNumberSchema()
])

AUGMENTATION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Path=PATH_SCHEMA,
        Function=IDENTIFIER_SCHEMA,
        Class=IDENTIFIER_SCHEMA,
        InputParameters=JsonObjectSchema(
            additional_properties=True,
        ),
    ),
    required=[
        'Path',
        'Function',
    ],
    additional_properties=False,
)

ACCESS_CONTROL_SCHEMA = JsonObjectSchema(
    properties=dict(
        IsSubstitute=BOOLEAN_SCHEMA,
        RequiredScopes=JsonArraySchema(items=IDENTIFIER_SCHEMA)
    ),
    additional_properties=False,
)

COMMON_DATASET_PROPERTIES = dict(
    Title=STRING_SCHEMA,
    Variables=VARIABLES_SCHEMA,
    TimeSeriesDataset=IDENTIFIER_SCHEMA,
    BoundingBox=GEO_BOUNDING_BOX_SCHEMA,
    ChunkCacheSize=CHUNK_SIZE_SCHEMA,
    Augmentation=AUGMENTATION_SCHEMA,
    Style=IDENTIFIER_SCHEMA,
    Hidden=BOOLEAN_SCHEMA,
    AccessControl=ACCESS_CONTROL_SCHEMA,
    PlaceGroups=JsonArraySchema(items=PLACE_GROUP_SCHEMA),
    Attribution=ATTRIBUTION_SCHEMA
)

DATASET_CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA,
        StoreInstanceId=IDENTIFIER_SCHEMA,  # will be set by server
        Path=PATH_SCHEMA,
        FileSystem=FILE_SYSTEM_SCHEMA,
        Anonymous=BOOLEAN_SCHEMA,
        Endpoint=URI_SCHEMA,
        Region=IDENTIFIER_SCHEMA,
        Function=IDENTIFIER_SCHEMA,
        Class=IDENTIFIER_SCHEMA,
        InputDatasets=JsonArraySchema(items=IDENTIFIER_SCHEMA),
        InputParameters=JsonObjectSchema(additional_properties=True),
        **COMMON_DATASET_PROPERTIES,
    ),
    required=[
        'Identifier',
        'Path'
    ],
    additional_properties=False,
)

DATA_STORE_DATASET_SCHEMA = JsonObjectSchema(
    required=[
        'Path'
    ],
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA,
        Path=PATH_SCHEMA,
        StoreInstanceId=IDENTIFIER_SCHEMA,  # will be set by server
        StoreOpenParams=JsonObjectSchema(additional_properties=True),
        **COMMON_DATASET_PROPERTIES
    ),
    additional_properties=False,
)

DATA_STORE_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA,
        StoreId=IDENTIFIER_SCHEMA,
        StoreParams=JsonObjectSchema(additional_properties=True),
        Datasets=JsonArraySchema(items=DATA_STORE_DATASET_SCHEMA),
    ),
    required=[
        'Identifier',
        'StoreId',
    ],
    additional_properties=False,
)

COLOR_MAPPING_EXPLICIT_SCHEMA = JsonObjectSchema(
    properties=dict(
        ColorBar=STRING_SCHEMA,
        ValueRange=VALUE_RANGE_SCHEMA
    ),
    required=[],
    additional_properties=False
)

COLOR_MAPPING_BY_PATH_SCHEMA = JsonObjectSchema(
    properties=dict(
        ColorFile=STRING_SCHEMA,
    ),
    required=[
        "ColorFile",
    ],
    additional_properties=False
)

COLOR_MAPPING_SCHEMA = JsonComplexSchema(one_of=[
    COLOR_MAPPING_EXPLICIT_SCHEMA,
    COLOR_MAPPING_BY_PATH_SCHEMA,
])

CHANNEL_MAPPING_SCHEMA = JsonObjectSchema(
    properties=dict(
        ValueRange=JsonArraySchema(items=[
            JsonNumberSchema(),
            JsonNumberSchema()
        ]),
        Variable=STRING_SCHEMA
    ),
    required=[
        "ValueRange",
        "Variable"
    ],
    additional_properties=False
)

RGB_MAPPING_SCHEMA = JsonObjectSchema(
    properties=dict(
        Red=CHANNEL_MAPPING_SCHEMA,
        Green=CHANNEL_MAPPING_SCHEMA,
        Blue=CHANNEL_MAPPING_SCHEMA,
    ),
    required=[
        "Red",
        "Green",
        "Blue",
    ],
    additional_properties=False
)

STYLE_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=STRING_SCHEMA,
        ColorMappings=JsonObjectSchema(
            properties=dict(
                rgb=RGB_MAPPING_SCHEMA
            ),
            additional_properties=COLOR_MAPPING_SCHEMA
        )
    ),
    required=[
        'Identifier',
        'ColorMappings',
    ],
    additional_properties=False
)

SERVICE_PROVIDER_SCHEMA = JsonObjectSchema(
    additional_properties=True,
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        DatasetAttribution=ATTRIBUTION_SCHEMA,
        AccessControl=ACCESS_CONTROL_SCHEMA,
        DatasetChunkCacheSize=CHUNK_SIZE_SCHEMA,
        Datasets=JsonArraySchema(items=DATASET_CONFIG_SCHEMA),
        DataStores=JsonArraySchema(items=DATA_STORE_SCHEMA),
        Styles=JsonArraySchema(items=STYLE_SCHEMA),
        ServiceProvider=SERVICE_PROVIDER_SCHEMA,
    )
)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Tuple, Type, TypeVar, Union

from xcube.util.assertions import assert_true

T = TypeVar('T')

ItemType = Union[Type[T], Tuple[Type[T], ...]]
Pair = Tuple[T, T]
ScalarOrPair = Union[T, Pair]


def normalize_scalar_or_pair(
        value: ScalarOrPair[T],
        *,
        item_type: Optional[ItemType[T]] = None,
        name: Optional[str] = None
) -> Pair:
    try:
        assert_true(len(value) <= 2,
                    message=f"{name or 'Value'} must be a scalar or pair of "
                            f"{item_type or 'scalars'}, was '{value}'")
        x, y = value
    except TypeError:
        x, y = value, value
    if item_type is not None:
        assert_true(isinstance(x, item_type) and isinstance(y, item_type),
                    message=f"{name or 'Value'} must be a scalar or pair of "
                            f"{item_type}, was '{value}'")
    return x, y


# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Tuple, Union

import dask.array as da
import pyproj
import xarray as xr

from xcube.util.assertions import assert_true
from .base import GridMapping
from .helpers import _default_xy_dim_names
from .helpers import _default_xy_var_names
from .helpers import _normalize_crs
from .helpers import _normalize_int_pair
from .helpers import _normalize_number_pair
from .helpers import _to_int_or_float


class RegularGridMapping(GridMapping):

    def __init__(self, **kwargs):
        kwargs.pop('is_regular', None)
        super().__init__(is_regular=True, **kwargs)
        self._xy_coords = None

    def _new_x_coords(self) -> xr.DataArray:
        self._assert_regular()
        x_res = self.x_res
        x1, x2 = self.x_min + x_res / 2, self.x_max - x_res / 2
        x_name, _ = self.xy_dim_names
        return xr.DataArray(da.linspace(x1, x2,
                                        self.width,
                                        chunks=self.tile_width),
                            dims=self.xy_dim_names[0])

    def _new_y_coords(self) -> xr.DataArray:
        self._assert_regular()
        y_res = self.y_res
        y1, y2 = self.y_min + y_res / 2, self.y_max - y_res / 2
        if not self.is_j_axis_up:
            y1, y2 = y2, y1
        return xr.DataArray(da.linspace(y1, y2,
                                        self.height,
                                        chunks=self.tile_height),
                            dims=self.xy_dim_names[1])

    def _new_xy_coords(self) -> xr.DataArray:
        self._assert_regular()
        x_coords_1d = self.x_coords
        y_coords_1d = self.y_coords
        y_coords_2d, x_coords_2d = xr.broadcast(y_coords_1d, x_coords_1d)
        xy_coords = xr.concat([x_coords_2d, y_coords_2d],
                              dim='coord').chunk((2,
                                                  self.tile_height,
                                                  self.tile_width))
        xy_coords.name = 'xy_coords'
        return xy_coords


def new_regular_grid_mapping(
        size: Union[int, Tuple[int, int]],
        xy_min: Tuple[float, float],
        xy_res: Union[float, Tuple[float, float]],
        crs: Union[str, pyproj.crs.CRS],
        *,
        tile_size: Union[int, Tuple[int, int]] = None,
        is_j_axis_up: bool = False
) -> GridMapping:
    width, height = _normalize_int_pair(size, name='size')
    assert_true(width > 1 and height > 1, 'invalid size')

    x_min, y_min = _normalize_number_pair(xy_min, name='xy_min')

    x_res, y_res = _normalize_number_pair(xy_res, name='xy_res')
    assert_true(x_res > 0 and y_res > 0, 'invalid xy_res')

    crs = _normalize_crs(crs)

    x_min = _to_int_or_float(x_min)
    y_min = _to_int_or_float(y_min)
    x_max = _to_int_or_float(x_min + x_res * width)
    y_max = _to_int_or_float(y_min + y_res * height)

    if crs.is_geographic:
        # TODO: don't do that.
        #  Instead set NaN in coord vars returned by to_coords()
        if y_min < -90:
            raise ValueError('invalid y_min')
        if y_max > 90:
            raise ValueError('invalid size, y_min combination')

    return RegularGridMapping(
        crs=crs,
        size=(width, height),
        tile_size=tile_size or (width, height),
        xy_bbox=(x_min, y_min, x_max, y_max),
        xy_res=(x_res, y_res),
        xy_var_names=_default_xy_var_names(crs),
        xy_dim_names=_default_xy_dim_names(crs),
        is_lon_360=x_max > 180,
        is_j_axis_up=is_j_axis_up
    )


def to_regular_grid_mapping(grid_mapping: GridMapping,
                            *,
                            tile_size: Union[int, Tuple[int, int]] = None,
                            is_j_axis_up: bool = False) -> GridMapping:
    if grid_mapping.is_regular:
        if tile_size is not None \
                or is_j_axis_up != grid_mapping.is_j_axis_up:
            return grid_mapping.derive(tile_size=tile_size,
                                       is_j_axis_up=is_j_axis_up)
        return grid_mapping

    x_min, y_min, x_max, y_max = grid_mapping.xy_bbox
    x_res, y_res = grid_mapping.xy_res
    # x_digits = 2 + abs(round(math.log10(x_res)))
    # y_digits = 2 + abs(round(math.log10(y_res)))
    # x_min = floor_to_fraction(x_min, x_digits)
    # y_min = floor_to_fraction(y_min, y_digits)
    # x_max = ceil_to_fraction(x_max, x_digits)
    # y_max = ceil_to_fraction(y_max, y_digits)
    xy_res = min(x_res, y_res) or max(x_res, y_res)
    width = round((x_max - x_min + xy_res) / xy_res)
    height = round((y_max - y_min + xy_res) / xy_res)
    width = width if width >= 2 else 2
    height = height if height >= 2 else 2

    return new_regular_grid_mapping(
        size=(width, height),
        xy_min=(x_min, y_min),
        xy_res=xy_res,
        crs=grid_mapping.crs,
        tile_size=tile_size,
        is_j_axis_up=is_j_axis_up
    )



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os.path
import sys
from typing import Optional, Dict, Any, Sequence, Union

import jsonschema
import yaml

from xcube.core.byoa import CodeConfig
from xcube.util.assertions import assert_false
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from .config import CallbackConfig
from .config import CubeConfig
from .config import InputConfig
from .config import OutputConfig
from .error import CubeGeneratorError
from ...constants import LOG

CubeGeneratorRequestLike = Union[str, Dict, 'CubeGeneratorRequest']


class CubeGeneratorRequest(JsonObject):
    """
    A request used to generate data cubes using cube generators.

    :param input_config: A configuration for a single input.
        Must be omitted if *input_configs* is given.
    :param input_configs: A sequence of one or more input configurations.
        Must be omitted if *input_config* is given.
    :param cube_config: The target cube configuration.
    :param code_config: The user-code configuration.
    :param output_config: The output configuration for the target cube.
    :param callback_config: A configuration that allows a cube generator
        to publish progress information to a compatible endpoint.
    """

    def __init__(self,
                 input_config: InputConfig = None,
                 input_configs: Sequence[InputConfig] = None,
                 cube_config: CubeConfig = None,
                 code_config: CodeConfig = None,
                 output_config: OutputConfig = None,
                 callback_config: Optional[CallbackConfig] = None):
        assert_true(input_config or input_configs,
                    'one of input_config and input_configs must be given')
        assert_false(input_config and input_configs,
                     'input_config and input_configs cannot be given both')
        if input_config is not None:
            assert_instance(input_config, InputConfig, 'input_config')
            input_configs = [input_config]
        elif input_configs is not None:
            assert_instance(input_configs, (list, tuple), 'input_configs')
            for i in range(len(input_configs)):
                assert_instance(input_configs[i], InputConfig,
                                f'input_configs[{i}]')
        if cube_config is not None:
            assert_instance(cube_config, CubeConfig, 'cube_config')
        if code_config is not None:
            assert_instance(code_config, CodeConfig, 'code_config')
        assert_instance(output_config, OutputConfig, 'output_config')
        if callback_config is not None:
            assert_instance(callback_config, CallbackConfig, 'callback_config')
        self.input_configs = input_configs
        self.cube_config = cube_config
        self.code_config = code_config
        self.output_config = output_config
        self.callback_config = callback_config

    def for_service(self) -> 'CubeGeneratorRequest':
        if self.code_config is None:
            return self
        return CubeGeneratorRequest(
            input_configs=self.input_configs,
            cube_config=self.cube_config,
            code_config=self.code_config.for_service(),
            output_config=self.output_config,
            callback_config=self.callback_config,
        )

    def for_local(self) -> 'CubeGeneratorRequest':
        if self.code_config is None:
            return self
        return CubeGeneratorRequest(
            input_configs=self.input_configs,
            cube_config=self.cube_config,
            code_config=self.code_config.for_local(),
            output_config=self.output_config,
            callback_config=self.callback_config,
        )

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                input_config=InputConfig.get_schema(),
                input_configs=JsonArraySchema(
                    items=InputConfig.get_schema(), min_items=1
                ),
                cube_config=CubeConfig.get_schema(),
                code_config=CodeConfig.get_schema(),
                output_config=OutputConfig.get_schema(),
                callback_config=CallbackConfig.get_schema()
            ),
            required=['output_config'],
            factory=cls,
        )

    @classmethod
    def normalize(cls, request: CubeGeneratorRequestLike) \
            -> 'CubeGeneratorRequest':
        """
        Normalize given *request* to an instance of
        :class:CubeGeneratorRequest.

        If *request* is already a CubeGeneratorRequest it is returned as is.
        If it is a ``str``, it is interpreted as a YAML or JSON file path
        and the request is read from file using
        ``CubeGeneratorRequest.from_file()``.
        If it is a ``dict``, it is interpreted as a JSON object and the
        request is parsed using ``CubeGeneratorRequest.from_dict()``.

        :param request The request, or request file path,
            or request JSON object.
        :raise TypeError if *request* is not a ``CubeGeneratorRequest``,
            ``str``, or ``dict``.
        """
        if isinstance(request, CubeGeneratorRequest):
            return request
        if isinstance(request, str):
            return CubeGeneratorRequest.from_file(request)
        if isinstance(request, dict):
            return CubeGeneratorRequest.from_dict(request)
        raise TypeError('request must be a str, dict, '
                        'or a CubeGeneratorRequest instance')

    def to_dict(self) -> Dict[str, Any]:
        """Convert into a JSON-serializable dictionary"""
        d = {}

        if self.input_configs is not None:
            if len(self.input_configs) == 1:
                d.update(input_config=self.input_configs[0].to_dict())
            else:
                d.update(input_configs=[ic.to_dict()
                                        for ic in self.input_configs])

        if self.cube_config is not None:
            d.update(cube_config=self.cube_config.to_dict())

        if self.code_config is not None:
            d.update(code_config=self.code_config.to_dict())

        d.update(output_config=self.output_config.to_dict())

        if self.callback_config is not None:
            d.update(callback_config=self.callback_config.to_dict())

        return d

    @classmethod
    def from_dict(cls, request_dict: Dict[str, Any]) \
            -> 'CubeGeneratorRequest':
        """Create new instance from a JSON-serializable dictionary"""
        try:
            return cls.get_schema().from_instance(request_dict)
        except jsonschema.exceptions.ValidationError as e:
            raise CubeGeneratorError(f'{e}', status_code=400) from e

    @classmethod
    def from_file(cls, request_file: Optional[str], verbosity: int = 0) \
            -> 'CubeGeneratorRequest':
        """
        Create new instance from a JSON file, or YAML file,
        or JSON passed via stdin.
        """
        request_dict = cls._load_request_file(request_file,
                                              verbosity=verbosity)
        if verbosity:
            LOG.info(f'Cube generator request loaded '
                     f'from {request_file or "TTY"}.')
        return cls.from_dict(request_dict)

    @classmethod
    def _load_request_file(cls,
                           gen_config_file: Optional[str],
                           verbosity: int = 0) -> Dict:

        if gen_config_file is not None \
                and not os.path.exists(gen_config_file):
            raise CubeGeneratorError(f'Cube generator request '
                                     f'"{gen_config_file}" not found.')

        try:
            if gen_config_file is None:
                if not sys.stdin.isatty():
                    if verbosity:
                        LOG.info('Awaiting generator'
                                 ' request JSON from TTY...')
                    return json.load(sys.stdin)
            else:
                with open(gen_config_file, 'r') as fp:
                    if gen_config_file.endswith('.json'):
                        return json.load(fp)
                    else:
                        return yaml.safe_load(fp)
        except BaseException as e:
            raise CubeGeneratorError(f'Error loading generator request'
                                     f' "{gen_config_file}": {e}') from e

        raise CubeGeneratorError(f'Missing cube generator request.')



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional

import os.path

import click

YAML_FORMAT = "yaml"
JSON_FORMAT = "json"
DEFAULT_FORMAT = YAML_FORMAT


# noinspection PyShadowingBuiltins
@click.command(name="versions")
@click.option('--format', '-f', 'format_name',
              metavar='FORMAT',
              type=click.Choice([YAML_FORMAT, JSON_FORMAT],
                                case_sensitive=False),
              help=f'Output format. '
                   f'Must be one of {YAML_FORMAT!r} or {JSON_FORMAT!r}. '
                   f'If not given, derived from OUTPUT name extension.')
@click.option('--output', '-o', 'output_path',
              metavar='OUTPUT',
              help=f'Output file path.')
def versions(format_name: Optional[str], output_path: Optional[str]):
    """
    Get versions of important packages used by xcube.
    """
    from xcube.util.versions import get_xcube_versions
    xcube_versions = get_xcube_versions()

    if not format_name:
        format_name = YAML_FORMAT
        if output_path:
            _, ext = os.path.splitext(output_path)
            if ext.lower() == '.json':
                format_name = JSON_FORMAT

    if format_name.lower() == JSON_FORMAT:
        import json
        text = json.dumps(xcube_versions, indent=2)
    else:
        import yaml
        text = yaml.dump(xcube_versions)

    if output_path:
        with open(output_path, 'w') as fp:
            fp.write(text)
    else:
        print(text)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
import concurrent.futures
import copy
from typing import (Optional, Dict, Any, Union,
                    Callable, Sequence, Awaitable, Tuple, Type, List, Mapping)

import jsonschema.exceptions

from xcube.constants import EXTENSION_POINT_SERVER_APIS
from xcube.constants import LOG
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_subclass
from xcube.util.extension import ExtensionRegistry
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.plugin import get_extension_registry
from xcube.version import version
from .api import Api
from .api import ApiContext
from .api import ApiContextT
from .api import ApiRoute
from .api import ApiStaticRoute
from .api import Context
from .api import ReturnT
from .api import ServerConfig
from .asyncexec import AsyncExecution
from .config import BASE_SERVER_CONFIG_SCHEMA
from .config import get_url_prefix
from .config import resolve_config_path
from .framework import Framework
from ..util.frozen import FrozenDict


class Server(AsyncExecution):
    """
    A REST server extendable by API extensions.

    APIs are registered using the extension point
    "xcube.server.api".

    TODO:
      * Allow server to serve generic static content,
        e.g. "http://localhost:8080/images/outside-cube/${ID}.jpg"
      * Allow server updates triggered by local file changes
        and changes in s3 buckets
      * Address common server configuration
        - Configure server types by meta-configuration,
          e.g. server name, server description, names of APIs served,
          aliases for common server config...
        - Why we need aliases for common server config:
          o Camel case vs snake case parameters names,
            e.g. "BaseDir" vs "base_dir"
          o First capital letter in parameter names,
            e.g. "Address" vs "address"
      * Use any given request JSON schema in openAPI
        to validate requests in HTTP methods

    :param framework: The web server framework to be used
    :param config: The server configuration.
    :param extension_registry: Optional extension registry.
        Defaults to xcube's default extension registry.
    """

    def __init__(
            self,
            framework: Framework,
            config: Mapping[str, Any],
            extension_registry: Optional[ExtensionRegistry] = None,
    ):
        assert_instance(framework, Framework)
        assert_instance(config, collections.abc.Mapping)
        apis = self.load_apis(config,
                              extension_registry=extension_registry)
        for api in apis:
            LOG.info(f'Loaded service API {api.name!r}')
        static_routes = self._collect_static_routes(config)
        static_routes.extend(self._collect_api_static_routes(apis))
        routes = self._collect_api_routes(apis)
        url_prefix = get_url_prefix(config)
        framework.add_routes(routes, url_prefix)
        framework.add_static_routes(static_routes, url_prefix)
        self._framework = framework
        self._apis = apis
        self._config_schema = self.get_effective_config_schema(
            framework,
            apis
        )
        ctx = self._new_ctx(config)
        ctx.on_update(None)
        self._set_ctx(ctx)

    @property
    def framework(self) -> Framework:
        """The web server framework used by this server."""
        return self._framework

    @property
    def apis(self) -> Tuple[Api]:
        """The APIs supported by this server."""
        return self._apis

    @property
    def config_schema(self) -> JsonObjectSchema:
        """The effective JSON schema for the server configuration."""
        return self._config_schema

    @property
    def ctx(self) -> "ServerContext":
        """The current server context."""
        return self._ctx

    def _set_ctx(self, ctx: "ServerContext"):
        self._ctx = ctx
        self._framework.update(ctx)

    def _new_ctx(self, config: collections.abc.Mapping) -> "ServerContext":
        config = dict(config)
        for key in tuple(config.keys()):
            if key not in self._config_schema.properties:
                LOG.warning(f'Configuration setting {key!r} ignored,'
                            f' because there is no schema describing it.')
                config.pop(key)
        try:
            validated_config = self._config_schema.from_instance(config)
        except jsonschema.exceptions.ValidationError as e:
            raise ValueError(f"Invalid server configuration:\n{e}") from e
        return ServerContext(self, validated_config)

    def start(self):
        """Start this server."""
        LOG.info(f'Starting service...')
        for api in self._apis:
            api.on_start(self.ctx)
        self._framework.start(self.ctx)

    def stop(self):
        """Stop this server."""
        LOG.info(f'Stopping service...')
        self._framework.stop(self.ctx)
        for api in self._apis:
            api.on_stop(self.ctx)
        self._ctx.on_dispose()

    def update(self, config: Mapping[str, Any]):
        """Update this server with given server configuration."""
        ctx = self._new_ctx(config)
        ctx.on_update(prev_ctx=self._ctx)
        self._set_ctx(ctx)

    def call_later(self,
                   delay: Union[int, float],
                   callback: Callable,
                   *args,
                   **kwargs):
        """
        Executes the given callable *callback* after *delay* seconds.

        :param delay: Delay in seconds.
        :param callback: Callback to be called.
        :param args: Positional arguments passed to *callback*.
        :param kwargs: Keyword arguments passed to *callback*.
        """
        return self._framework.call_later(
            delay, callback, *args, **kwargs
        )

    def run_in_executor(
            self,
            executor: Optional[concurrent.futures.Executor],
            function: Callable[..., ReturnT],
            *args: Any,
            **kwargs: Any
    ) -> Awaitable[ReturnT]:
        """
        Concurrently runs a *function* in a ``concurrent.futures.Executor``.
        If *executor* is ``None``, the framework's default
        executor will be used.

        :param executor: An optional executor.
        :param function: The function to be run concurrently.
        :param args: Positional arguments passed to *function*.
        :param kwargs: Keyword arguments passed to *function*.
        :return: The awaitable return value of *function*.
        """
        return self._framework.run_in_executor(
            executor, function, *args, **kwargs
        )

    @classmethod
    def load_apis(
            cls,
            config: collections.abc.Mapping,
            extension_registry: Optional[ExtensionRegistry] = None
    ) -> Tuple[Api]:
        # Collect all registered API extensions
        extension_registry = extension_registry \
                             or get_extension_registry()
        api_extensions = extension_registry.find_extensions(
            EXTENSION_POINT_SERVER_APIS
        )

        # Get APIs specification
        api_spec = config.get("api_spec", {})
        incl_api_names = api_spec.get("includes",
                                      [ext.name for ext in api_extensions])
        excl_api_names = api_spec.get("excludes",
                                      [])

        # Collect effective APIs
        api_names = set(incl_api_names).difference(set(excl_api_names))
        apis: List[Api] = [ext.component
                           for ext in api_extensions
                           if ext.name in api_names]

        api_lookup = {api.name: api for api in apis}

        def assert_required_apis_available():
            # Assert that required APIs are available.
            for api in apis:
                for req_api_name in api.required_apis:
                    if req_api_name not in api_lookup:
                        raise ValueError(f'API {api.name!r}: missing API'
                                         f' dependency {req_api_name!r}')

        assert_required_apis_available()

        def count_api_refs(api: Api) -> int:
            # Count the number of times the given API is referenced.
            dep_sum = 0
            for req_api_name in api.required_apis:
                dep_sum += count_api_refs(api_lookup[req_api_name]) + 1
            for opt_api_name in api.optional_apis:
                if opt_api_name in api_lookup:
                    dep_sum += count_api_refs(api_lookup[opt_api_name]) + 1
            return dep_sum

        # Count the number of times each API is referenced.
        api_ref_counts = {
            api.name: count_api_refs(api)
            for api in apis
        }

        # Return an ordered dict sorted by an API's reference count
        return tuple(sorted(apis,
                            key=lambda api: api_ref_counts[api.name]))

    @classmethod
    def _collect_static_routes(cls, config: collections.abc.Mapping) \
            -> List[ApiStaticRoute]:
        static_routes = config.get('static_routes', [])
        api_static_routes = []
        for static_route in static_routes:
            params = dict(**static_route)
            dir_path = params.get("dir_path")
            if dir_path is not None:
                dir_path = resolve_config_path(config, dir_path)
                params["dir_path"] = dir_path
            try:
                api_static_route = ApiStaticRoute(**params)
                api_static_routes.append(api_static_route)
            except (TypeError, ValueError):
                LOG.error(f"Failed to add static route: {params!r}",
                          exc_info=True)
        return api_static_routes

    @classmethod
    def _collect_api_static_routes(cls, apis: Sequence[Api]) \
            -> List[ApiStaticRoute]:
        static_routes = []
        for api in apis:
            static_routes.extend(api.static_routes)
        return static_routes

    @classmethod
    def _collect_api_routes(cls, apis: Sequence[Api]) -> List[ApiRoute]:
        handlers = []
        for api in apis:
            handlers.extend(api.routes)
        return handlers

    @classmethod
    def get_effective_config_schema(
            cls,
            framework: Framework,
            apis: Sequence[Api]
    ) -> JsonObjectSchema:
        effective_config_schema = copy.deepcopy(BASE_SERVER_CONFIG_SCHEMA)
        framework_config_schema = framework.config_schema
        if framework_config_schema is not None:
            cls._update_config_schema(effective_config_schema,
                                      framework_config_schema,
                                      f'Server')
        for api in apis:
            api_config_schema = api.config_schema
            if api_config_schema is not None:
                cls._update_config_schema(effective_config_schema,
                                          api_config_schema,
                                          f'API {api.name!r}')
        return effective_config_schema

    @classmethod
    def _update_config_schema(cls,
                              config_schema: JsonObjectSchema,
                              config_schema_update: JsonObjectSchema,
                              schema_name: str):
        assert isinstance(config_schema, JsonObjectSchema)
        assert isinstance(config_schema_update, JsonObjectSchema)
        for k, v in config_schema_update.properties.items():
            if k in config_schema.properties:
                raise ValueError(f'{schema_name}:'
                                 f' configuration parameter {k!r}'
                                 f' is already defined.')
            config_schema.properties[k] = v
        if config_schema_update.required:
            for r in config_schema_update.required:
                if r not in config_schema.required:
                    config_schema.required.append(r)

    def get_open_api_doc(self, include_all: bool = False) -> Dict[str, Any]:
        """Get the OpenAPI JSON document for this server."""
        error_schema = {
            "type": "object",
            "properties": {
                "status_code": {
                    "type": "integer",
                    "minimum": 200,
                },
                "message": {
                    "type": "string",
                },
                "reason": {
                    "type": "string",
                },
                "exception": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    }
                }
            },
            "additionalProperties": True,
            "required": ["status_code", "message"],
        }

        schema_components = {
            "Error": {
                "type": "object",
                "properties": {
                    "error": error_schema,
                },
                "additionalProperties": True,
                "required": ["error"],
            }
        }

        response_components = {
            "UnexpectedError": {
                "description": "Unexpected error.",
                "content": {
                    "application/json": {
                        "schema": {
                            "$ref": "#/components/schemas/Error"
                        }
                    }
                }
            }
        }

        default_responses = {
            "200": {
                "description": "On success.",
            },
            "default": {
                "$ref": "#/components/responses/UnexpectedError"
            }
        }

        url_prefix = get_url_prefix(self.ctx.config)

        tags = []
        paths = {}
        for api in self.ctx.apis:
            if not api.routes and not api.static_routes:
                # Only include APIs with endpoints
                continue
            tags.append({
                "name": api.name,
                "description": api.description or ""
            })
            for route in api.routes:
                if not include_all \
                        and route.path.startswith('/maintenance/'):
                    continue
                path = dict(
                    description=getattr(
                        route.handler_cls, "__doc__", ""
                    ) or ""
                )
                for method in ("head",
                               "get",
                               "post",
                               "put",
                               "delete",
                               "options"):
                    fn = getattr(route.handler_cls, method, None)
                    openapi_metadata = getattr(fn, '__openapi__', None)
                    if isinstance(openapi_metadata, dict):
                        openapi_metadata = openapi_metadata.copy()
                        if 'tags' not in openapi_metadata:
                            openapi_metadata['tags'] = [api.name]
                        if 'description' not in openapi_metadata:
                            openapi_metadata['description'] = \
                                getattr(fn, "__doc__", None) or ""
                        responses = openapi_metadata.get("responses")
                        if responses is None:
                            responses = default_responses.copy()
                        elif isinstance(responses, dict):
                            _responses = default_responses.copy()
                            _responses.update(responses)
                            responses = _responses
                        openapi_metadata['responses'] = responses
                        path[method] = dict(**openapi_metadata)
                paths[route.path] = path
            for route in api.static_routes:
                openapi_metadata = dict(route.openapi_metadata or {})
                if 'tags' not in openapi_metadata:
                    openapi_metadata['tags'] = [api.name]
                paths[route.path] = dict(
                    get=dict(**openapi_metadata)
                )

        return {
            "openapi": "3.0.0",
            "info": {
                "title": "xcube Server",
                "description": "xcube Server API",
                "version": version,
            },
            "servers": [
                {
                    # TODO (forman): the following URL must be adjusted
                    #   e.g. pass request.url_for_path('') as url into
                    #   this method, or even pass the list of servers.
                    "url": f"http://localhost:8080{url_prefix}",
                    "description": "Local development server."
                },
            ],
            "tags": tags,
            "paths": paths,
            "components": {
                "schemas": schema_components,
                "responses": response_components
            }
        }


class ServerContext(Context):
    """
    The server context holds the current server configuration and
    the API context objects that depend on that specific configuration.

    A new server context is created for any new server configuration,
    which in turn will cause all API context objects to be updated.

    The constructor shall not be called directly.

    :param server: The server.
    :param config: The current server configuration.
    """

    def __init__(self,
                 server: Server,
                 config: collections.abc.Mapping):
        self._server = server
        self._config = FrozenDict.freeze(config)
        self._api_contexts: Dict[str, Context] = dict()

    @property
    def server(self) -> Server:
        return self._server

    @property
    def apis(self) -> Tuple[Api]:
        return self._server.apis

    def get_open_api_doc(self, include_all: bool = False) -> Dict[str, Any]:
        return self._server.get_open_api_doc(include_all=include_all)

    @property
    def config(self) -> ServerConfig:
        return self._config

    def get_api_ctx(self,
                    api_name: str,
                    cls: Optional[Type[ApiContextT]] = None) \
            -> Optional[ApiContextT]:
        api_ctx = self._api_contexts.get(api_name)
        if cls is not None:
            assert_subclass(cls, ApiContext, name='cls')
            assert_instance(api_ctx, cls,
                            name=f'api_ctx (context of API {api_name!r})')
        return api_ctx

    def _set_api_ctx(self, api_name: str, api_ctx: ApiContext):
        assert_instance(api_ctx, ApiContext,
                        name=f'api_ctx (context of API {api_name!r})')
        self._api_contexts[api_name] = api_ctx
        setattr(self, api_name, api_ctx)

    def call_later(self,
                   delay: Union[int, float],
                   callback: Callable,
                   *args,
                   **kwargs) -> object:
        return self._server.call_later(delay, callback,
                                       *args, **kwargs)

    def run_in_executor(self,
                        executor: Optional[concurrent.futures.Executor],
                        function: Callable[..., ReturnT],
                        *args: Any,
                        **kwargs: Any) -> Awaitable[ReturnT]:
        return self._server.run_in_executor(executor, function,
                                            *args, **kwargs)

    def on_update(self, prev_ctx: Optional["ServerContext"]):
        if prev_ctx is None:
            LOG.info(f'Applying initial configuration...')
        else:
            LOG.info(f'Applying configuration changes...')
        for api in self.apis:
            prev_api_ctx: Optional[ApiContext] = None
            if prev_ctx is not None:
                prev_api_ctx = prev_ctx.get_api_ctx(api.name)
                assert prev_api_ctx is not None
            for dep_api_name in api.required_apis:
                dep_api_ctx = self.get_api_ctx(dep_api_name)
                assert dep_api_ctx is not None
            next_api_ctx: Optional[ApiContext] = api.create_ctx(self)
            self._set_api_ctx(api.name, next_api_ctx)
            next_api_ctx.on_update(prev_api_ctx)
            if prev_api_ctx is not None \
                    and prev_api_ctx is not next_api_ctx:
                prev_api_ctx.on_dispose()

    def on_dispose(self):
        for api_name in reversed([api.name for api in self.apis]):
            api_ctx = self.get_api_ctx(api_name)
            if api_ctx is not None:
                api_ctx.on_dispose()



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click


# noinspection PyShadowingBuiltins,PyUnusedLocal
@click.command(name="vars2dim")
@click.argument('cube')
@click.option('--variable', '--var', metavar='VARIABLE',
              default='data',
              help='Name of the new variable that includes all variables. Defaults to "data".')
@click.option('--dim_name', '-D', metavar='DIM-NAME',
              default='var',
              help='Name of the new dimension into variables. Defaults to "var".')
@click.option('--output', '-o', metavar='OUTPUT',
              help="Output path. If omitted, 'INPUT-vars2dim.FORMAT' will be used.")
@click.option('--format', '-f', metavar='FORMAT', type=click.Choice(['zarr', 'netcdf']),
              help="Format of the output. If not given, guessed from OUTPUT.")
def vars2dim(cube, variable, dim_name, output=None, format=None):
    """
    Convert cube variables into new dimension.
    Moves all variables of CUBE into a single new variable <var-name>
    with a new dimension DIM-NAME and writes the results to OUTPUT.
    """

    from xcube.core.dsio import guess_dataset_format
    from xcube.core.dsio import open_dataset, write_dataset
    from xcube.core.vars2dim import vars_to_dim
    import os

    if not output:
        dirname = os.path.dirname(cube)
        basename = os.path.basename(cube)
        basename, ext = os.path.splitext(basename)
        output = os.path.join(dirname, basename + '-vars2dim' + ext)

    format_name = format if format else guess_dataset_format(output)

    with open_dataset(input_path=cube) as ds:
        converted_dataset = vars_to_dim(ds, dim_name=dim_name, var_name=variable)
        write_dataset(converted_dataset, output_path=output, format_name=format_name)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import threading
from time import sleep
from timeit import default_timer
from typing import Sequence

import requests

from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_true
from xcube.util.progress import ProgressObserver
from xcube.util.progress import ProgressState
from .config import CallbackConfig
from ...constants import LOG


def _format_time(t):
    """Format seconds into a human readable form.
    Taken form Dask

    >>> _format_time(10.4)
    '10.4s'
    >>> _format_time(1000.4)
    '16min 40.4s'
    """
    m, s = divmod(t, 60)
    h, m = divmod(m, 60)
    if h:
        return "{0:2.0f}hr {1:2.0f}min {2:4.1f}s".format(h, m, s)
    elif m:
        return "{0:2.0f}min {1:4.1f}s".format(m, s)
    else:
        return "{0:4.1f}s".format(s)


class _ThreadedProgressObserver(ProgressObserver):
    """
    A threaded Progress observer adapted from Dask's ProgressBar class.
    """

    def __init__(self,
                 minimum: float = 0,
                 dt: float = 1,
                 timeout: float = None):
        """

        :type dt: float
        :type minimum: float
        """
        super().__init__()
        assert_true(dt >= 0, "The timer's time step must be >=0")
        assert_true(minimum >= 0, "The timer's minimum must be >=0")
        self._running = False
        self._start_time = None
        self._minimum = minimum
        self._dt = dt
        self._width = 100
        self._current_sender = None
        self._state_stack: [ProgressState] = []
        self._timeout = timeout

    def _timer_func(self):
        """Background thread for updating the progress bar"""
        while self._running:
            elapsed = default_timer() - self._start_time
            if elapsed > self._minimum:
                self._update_state(elapsed)
            if self._timeout and elapsed > self._timeout:
                self._running = False
            sleep(self._dt)

    def _start_timer(self):
        self._width = self._state_stack[0].total_work
        self._start_time = default_timer()
        # Start background thread
        self._running = True
        self._timer = threading.Thread(target=self._timer_func)
        self._timer.daemon = True
        self._timer.start()

    def _stop_timer(self, errored):
        self._running = False
        self._timer.join()
        elapsed = default_timer() - self._start_time
        self.last_duration = elapsed
        if elapsed < self._minimum:
            return
        if not errored:
            self.callback(self._current_sender, elapsed, self._state_stack)
        else:
            self._update_state(elapsed)

    def _update_state(self, elapsed):
        if not self._state_stack:
            self.callback(self._current_sender, 0, elapsed)
            return
        state = self._state_stack[0]

        if state.completed_work < state.total_work:
            self.callback(self._current_sender, elapsed, self._state_stack)

    def on_begin(self, state_stack: Sequence[ProgressState]):
        assert_given(state_stack, name='state_stack')
        self._state_stack = state_stack
        self._current_sender = "on_begin"
        if not self._running and len(state_stack) == 1:
            self._start_timer()

    def on_update(self, state_stack: Sequence[ProgressState]):
        assert_given(state_stack, name="state_stack")
        self._state_stack = state_stack
        self._current_sender = "on_update"

    def on_end(self, state_stack: Sequence[ProgressState]):
        assert_given(state_stack, name="state_stack")
        self._state_stack = state_stack
        self._current_sender = "on_end"

        if self._running and len(state_stack) == 1:
            self._stop_timer(False)

    def callback(self,
                 sender: str,
                 elapsed: float,
                 state_stack: Sequence[ProgressState]):
        """

        :param sender:
        :param elapsed:
        :param state_stack:
        :return:
        """


class ApiProgressCallbackObserver(_ThreadedProgressObserver):
    def __init__(self,
                 callback_config: CallbackConfig,
                 minimum: float = 0,
                 dt: float = 1,
                 timeout: float = False):
        super().__init__(minimum=minimum, dt=dt, timeout=timeout)
        assert_true(callback_config.api_uri and callback_config.access_token,
                    "Both, api_uri and access_token must be given.")

        self.callback_config = callback_config

    def callback(self,
                 sender: str,
                 elapsed: float,
                 state_stack: Sequence[ProgressState]):
        assert_given(state_stack, "ProgressStates")
        state = state_stack[0]
        callback = {
            "sender": sender,
            "state": {
                "label": state.label,
                "total_work": state.total_work,
                "error": state.exc_info_text or False,
                "progress": state.progress,
                "elapsed": elapsed,
            }
        }
        callback_api_uri = self.callback_config.api_uri
        callback_api_access_token = self.callback_config.access_token
        header = {"Authorization": f"Bearer {callback_api_access_token}"}

        return requests.put(callback_api_uri, json=callback, headers=header)


class TerminalProgressCallbackObserver(_ThreadedProgressObserver):
    def __init__(self,
                 minimum: float = 0,
                 dt: float = 1,
                 timeout: float = False):
        super().__init__(minimum=minimum, dt=dt, timeout=timeout)

    def callback(self,
                 sender: str,
                 elapsed: float,
                 state_stack: [ProgressState],
                 prt: bool = True):
        state = state_stack[0]

        bar = "#" * int(state.total_work * state.progress)
        percent = int(100 * state.progress)
        elapsed = _format_time(elapsed)
        msg = "\r{0}: [{1:<{2}}] | {3}% Completed | {4}".format(
            state.label, bar, state.total_work, percent, elapsed
        )

        if prt:
            LOG.info(msg)

        return msg


class ConsoleProgressObserver(ProgressObserver):

    def on_begin(self, state_stack: Sequence[ProgressState]):
        LOG.info(self._format_progress(state_stack,
                                       status_label='...'))

    def on_update(self, state_stack: Sequence[ProgressState]):
        LOG.info(self._format_progress(state_stack))

    def on_end(self, state_stack: Sequence[ProgressState]):
        if state_stack[0].exc_info:
            LOG.info(self._format_progress(state_stack,
                                           status_label='error!'))
        else:
            LOG.info(self._format_progress(state_stack,
                                           status_label='done.'))

    @classmethod
    def _format_progress(cls,
                         state_stack: Sequence[ProgressState],
                         status_label=None) -> str:
        if status_label:
            state_stack_part = cls._format_state_stack(state_stack[0:-1])
            state_part = cls._format_state(state_stack[-1],
                                           marker=status_label)
            return state_part \
                if not state_stack_part \
                else state_stack_part + ': ' + state_part
        else:
            return cls._format_state_stack(state_stack)

    @classmethod
    def _format_state_stack(cls,
                            state_stack: Sequence[ProgressState],
                            marker=None) -> str:
        return ': '.join([cls._format_state(s) for s in state_stack])

    @classmethod
    def _format_state(cls, state: ProgressState, marker=None) -> str:
        if marker is None:
            return '{} - {:3.1%}'.format(state.label, state.progress)
        else:
            return '{} - {}'.format(state.label, marker)



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import jsonschema
from xcube.server.api import ApiError
from xcube.server.api import ApiHandler
from .api import api
from .context import ComputeContext
from .controllers import get_compute_operations
from .controllers import get_compute_operation
from .context import JOB_STATUSES

OP_SCHEMA = {
    "type": "object",
    "properties": {
        "operationId": {"type": "string"},
        "parametersSchema": {
            "type": "array",
            "items": {
                "type": "object",
                "description": "JSON Schema for each parameter",
            }
        },
        "description": {"type": "string"},
    },
    "required": ["operationId", "parametersSchema"],
}

OP_LIST_SCHEMA = {
    "type": "object",
    "properties": {
        "operations": {
            "type": "array",
            "items": OP_SCHEMA
        },
    },
    "required": ["operations"],
}

OP_RESPONSES_SCHEMA = {
    "200": {
        "description": "Compute operation details",
        "content": {
            "application/json": OP_SCHEMA
        }
    },
    "404": {
        "description": "An operation with the specified ID was not found."
    }
}

OP_LIST_RESPONSES_SCHEMA = {
    "200": {
        "description": "Compute operation details",
        "content": {
            "application/json": {
                "schema": OP_LIST_SCHEMA
            }
        }
    },
    "404": {
        "description": "An operation with the specified ID was not found."
    }
}

JOB_STATE_SCHEMA = {
    "type": "object",
    "properties": {
        "status": {
            "type": "string",
            "enum": list(JOB_STATUSES)
        },
        "error": {
            "type": "object",
            "properties": {
                "message": {"type": "string"},
                "type": {"type": "string"},
                "traceback": {
                    "type": "array",
                    "items": {"type": "string"}
                },
            },
            "required": ["message"]
        },
    }
}

JOB_RESULT_SCHEMA = {
    "type": "object",
    "properties": {
        "datasetId": {"type": "string"},
        "data": {},
    }
}

JOB_SCHEMA = {
    "type": "object",
    "properties": {
        "jobId": {"type": "integer"},
        "state": JOB_STATE_SCHEMA,
        "result": JOB_RESULT_SCHEMA,
    },
    "required": ["jobId", "state"],
}

JOB_LIST_SCHEMA = {
    "type": "object",
    "properties": {
        "jobs": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "jobId": {"type": "integer"},
                    "operationId": {"type": "string"},
                    "status": {
                        "type": "string",
                        "enum": list(JOB_STATUSES)
                    },
                },
                "required": ["jobId", "operationId", "status"],
            }
        },
    },
    "required": ["jobs"],
}

JOB_RESPONSES_SCHEMA = {
    "200": {
        "description": "Compute job details",
        "content": {
            "application/json": JOB_SCHEMA
        }
    },
    "404": {
        "description": "A job with the specified ID was not found."
    }
}

JOB_LIST_RESPONSES_SCHEMA = {
    "200": {
        "description": "List of compute jobs",
        "content": {
            "application/json": {
                "schema": JOB_LIST_SCHEMA
            }
        }
    }
}

JOB_REQUEST_SCHEMA = {
    "type": "object",
    "properties": {
        "operationId": {
            "type": "string"
        },
        "parameters": {
            "type": "object",
            "items": {}  # Any
        },
        "output": {
            "type": "object",
            "properties": {
                "datasetId": {"type": "string"},
                "title": {"type": "string"},
                "description": {"type": "string"},
                "style": {"type": "string"},
                "colorMappings": {"type": "object"},
            },
            # Doesn't work with swagger:
            # TypeError: t.get is not a function
            # "additionalProperties": False,
        },
        # Doesn't work with swagger:
        # TypeError: t.get is not a function
        # "additionalProperties": False,
    },
    "required": ["operationId"],
    "example": {
        "operationId": "spatial_subset",
        "parameters": {
            "dataset": "demo",
            "bbox": [1, 51, 4, 52]
        },
        "output": {
            "datasetId": "demo_subset",
            "title": "My demo subset"
        }
    }
}

JOB_REQUEST_BODY_SCHEMA = {
    "description": "Compute job request",
    "required": True,
    "content": {
        "application/json": {
            "schema": JOB_REQUEST_SCHEMA
        }
    }
}

PATH_PARAM_OP_ID = {
    "name": "operationId",
    "in": "path",
    "description": "Operation identifier",
    "schema": {"type": "string"}
}

PATH_PARAM_JOB_ID = {
    "name": "jobId",
    "in": "path",
    "description": "Job identifier",
    "schema": {"type": "integer"}
}


@api.route('/compute/operations')
class ComputeOperationsHandler(ApiHandler[ComputeContext]):

    @api.operation(
        operation_id="getComputeOperations",
        summary="Get available compute operations.",
        responses=OP_LIST_RESPONSES_SCHEMA
    )
    def get(self):
        self.response.finish(get_compute_operations(self.ctx))


# noinspection PyPep8Naming
@api.route('/compute/operations/{operationId}')
class ComputeOperationHandler(ApiHandler[ComputeContext]):

    @api.operation(
        operation_id="getComputeOperationInfo",
        summary="Get the details of a given compute operation.",
        parameters=[PATH_PARAM_OP_ID],
        responses=OP_RESPONSES_SCHEMA
    )
    def get(self, operationId):
        self.response.finish(get_compute_operation(self.ctx, operationId))


# noinspection PyPep8Naming
@api.route('/compute/jobs')
class ComputeJobsHandler(ApiHandler[ComputeContext]):

    @api.operation(
        operation_id="getComputeJobs",
        summary="Get recent compute jobs.",
        responses=JOB_LIST_RESPONSES_SCHEMA,
    )
    def get(self):
        self.response.finish({
            "jobs": [
                {
                    "jobId": job["jobId"],
                    "operationId": job["request"]["operationId"],
                    "status": job["state"]["status"],
                }
                for job in self.ctx.jobs.values()
            ]
        })

    @api.operation(
        operation_id="scheduleComputeJob",
        summary="Schedule a new compute job.",
        request_body=JOB_REQUEST_BODY_SCHEMA,
        responses=JOB_RESPONSES_SCHEMA,
    )
    def put(self):
        # TODO: build a more complete schema and publish it as OpenAPI?
        #  Currently, the schema is built for the current request and only
        #  contains (and only needs to contain) a parameter schema for the
        #  requested operation. In xcube’s live OpenAPI specification,
        #  the parameters are just an object with no further constraints.
        #  Ideally we’d want to detail the parameters for the various
        #  available operations in the OpenAPI specification. This can be
        #  done with if-then constructs: see
        #  https://json-schema.org/understanding-json-schema/reference/conditionals.html#if-then-else
        job_request = self.request.json
        basic_schema = self.put.__openapi__['requestBody']['content'][
            'application/json']['schema']
        # noinspection PyProtectedMember,PyUnresolvedReferences
        op_schema = dict(
            properties=dict(
                operationId=dict(enum=list(self.ctx.op_registry.ops.keys())),
                parameters=self.ctx.op_registry.ops[
                    job_request['operationId']]._op_info.params_schema
            )
        )
        full_schema = {**basic_schema, **op_schema}
        try:
            jsonschema.validate(job_request, full_schema)
            self.response.finish(self.ctx.schedule_job(job_request))
        except jsonschema.ValidationError as e:
            raise ApiError.BadRequest(message=f'{e.message} at {e.json_path}')


# noinspection PyPep8Naming
@api.route('/compute/jobs/{jobId}')
class ComputeJobHandler(ApiHandler[ComputeContext]):

    @api.operation(
        operation_id="getComputeJob",
        summary="Get details about a compute job.",
        parameters=[PATH_PARAM_JOB_ID],
        responses=JOB_RESPONSES_SCHEMA,
    )
    def get(self, jobId: str):
        job = self.ctx.jobs.get(int(jobId))
        if job is None:
            raise ApiError.NotFound(f"job #{jobId} cannot be found")
        self.response.finish(job)

    @api.operation(
        operation_id="cancelComputeJob",
        summary="Cancel an existing compute job.",
        parameters=[PATH_PARAM_JOB_ID],
        responses=JOB_RESPONSES_SCHEMA,
    )
    def delete(self, jobId: str):
        self.response.finish(self.ctx.cancel_job(int(jobId)))



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click


# noinspection PyShadowingBuiltins
@click.command(name="dump")
@click.argument('input')
@click.option('--variable', '--var', metavar='VARIABLE', multiple=True,
              help="Name of a variable (multiple allowed).")
@click.option('--encoding', '-E', is_flag=True, flag_value=True,
              help="Dump also variable encoding information.")
def dump(input, variable, encoding):
    """
    Dump contents of an input dataset.
    """
    from xcube.core.dsio import open_dataset
    from xcube.core.dump import dump_dataset
    with open_dataset(input) as ds:
        text = dump_dataset(ds, var_names=variable, show_var_encoding=encoding)
        print(text)



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import json
import os
import socket
import threading
from pathlib import Path
from typing import Optional, Union, Mapping, Any, Tuple, Dict

import tornado.ioloop
import xarray as xr

from xcube.constants import LOG
from xcube.core.mldataset import MultiLevelDataset
from xcube.server.server import Server
from xcube.server.webservers.tornado import TornadoFramework
from xcube.webapi.datasets.context import DatasetsContext


# Name of the env var that contains a JupyterLab's base URL.
# If this env var is set, the following _LAB_INFO_FILE will not be used.
# We also imply that a jupyter-server-proxy is installed and enabled.
_LAB_URL_ENV_VAR = "XCUBE_JUPYTER_LAB_URL"

# The following file is generated by the xcube JupyterLab extension
# (xcube-jl-ext) if it is installed and enabled in JupyterLab.
_LAB_INFO_FILE = "~/.xcube/jupyterlab/lab-info.json"


class Viewer:
    """
    Experimental class that represents the xcube Viewer
    in Jupyter Notebooks.

    :param server_config: Server configuration.
        See "xcube serve --show configschema".
    """

    def __init__(self, server_config: Optional[Mapping[str, Any]] = None):
        server_config = dict(server_config or {})

        port = server_config.get("port")
        address = server_config.get("address")

        if port is None:
            port = _find_port()
        if address is None:
            address = "0.0.0.0"

        server_config["port"] = port
        server_config["address"] = address

        server_url, reverse_url_prefix = _get_server_url_and_rev_prefix(port)
        server_config["reverse_url_prefix"] = reverse_url_prefix

        self._server_config = server_config

        # Got trick from
        # https://stackoverflow.com/questions/55201748/running-a-tornado-server-within-a-jupyter-notebook
        self._io_loop = tornado.ioloop.IOLoop()
        thread = threading.Thread(target=self._io_loop.start)
        thread.daemon = True
        thread.start()

        self._server = Server(TornadoFramework(io_loop=self._io_loop,
                                               shared_io_loop=True),
                              config=server_config)

        self._io_loop.add_callback(self._server.start)

        self._server_url = server_url
        self._viewer_url = f"{server_url}/viewer/?serverUrl={server_url}"

    @property
    def server_config(self) -> Mapping[str, Any]:
        """The server configuration used by this viewer."""
        return self._server_config

    @property
    def server_url(self):
        """The URL of the server used by this viewer."""
        return self._server_url

    @property
    def viewer_url(self):
        """The URL of this viewer."""
        return self._viewer_url

    @property
    def is_server_running(self) -> bool:
        """Whether the server is running."""
        return self._server is not None

    @property
    def datasets_ctx(self) -> DatasetsContext:
        """Gets the context for the server's "datasets" API."""
        assert self.is_server_running
        return self._server.ctx.get_api_ctx('datasets')

    def stop_server(self):
        """Stops this viewer's server."""
        if self._server is not None:
            # noinspection PyBroadException
            try:
                self._server.stop()
            except BaseException:
                pass
        self._server = None
        self._io_loop = None

    def add_dataset(self,
                    dataset: Union[xr.Dataset, MultiLevelDataset],
                    ds_id: Optional[str] = None,
                    title: Optional[str] = None,
                    style: Optional[str] = None,
                    color_mappings: Dict[str, Dict[str, Any]] = None):
        """
        Add a dataset to this viewer.

        :param dataset: The dataset to me added. Must be an instance of
            ``xarray.Dataset`` or ``xcube.core.mldataset.MultiLevelDataset``.
        :param ds_id: Optional dataset identifier.
            If not given, an identifier will be generated and returned.
        :param title: Optional dataset title.
            Overrides a title given by dataset metadata.
        :param style: Optional name of a style that must exist
            in the server configuration.
        :param color_mappings: Maps a variable name to a specific
            color mapping that is a dictionary comprising a "ValueRange"
            (a pair of numbers) and a "ColorBar"
            (a matplotlib color bar name).
        :return: The dataset identifier.
        """
        if not self._check_server_running():
            return
        return self.datasets_ctx.add_dataset(dataset,
                                             ds_id=ds_id,
                                             title=title,
                                             style=style,
                                             color_mappings=color_mappings)

    def remove_dataset(self, ds_id: str):
        """
        Remove a dataset from this viewer.

        :param ds_id: The identifier of the dataset to be removed.
        """
        if not self._check_server_running():
            return
        self.datasets_ctx.remove_dataset(ds_id)

    def show(self,
             width: Union[int, str] = "100%",
             height: Union[str, int] = 800):
        """
        Show this viewer as an iframe.
        Intended to be used in a Jupyter notebook.
        If used outside a Jupyter notebook the viewer will be shown
        as a new browser tab.

        :param width: The width of the viewer's iframe.
        :param height: The height of the viewer's iframe.
        """
        try:
            from IPython.core.display import HTML
            return HTML(
                f'<iframe src="{self._viewer_url}&compact=1"'
                f' width="{width}"'
                f' height="{height}"'
                f'/>'
            )
        except ImportError as e:
            print(f"Error: {e}; Trying to open Viewer in web browser...")
            # noinspection PyBroadException
            try:
                import webbrowser
                webbrowser.open_new_tab(self.viewer_url)
            except BaseException:
                print("Failed too.")

    def info(self):
        """Output viewer info."""
        # Consider outputting this as HTML if in Notebook
        print(f"Server: {self.server_url}")
        print(f"Viewer: {self.viewer_url}")

    def _check_server_running(self):
        if not self.is_server_running:
            print('Server not running')
        return self.is_server_running


def _get_server_url_and_rev_prefix(port: int) -> Tuple[str, str]:
    lab_url = os.environ.get(_LAB_URL_ENV_VAR) or None
    has_proxy = lab_url is not None

    if not lab_url:
        lab_info_path = Path(*_LAB_INFO_FILE.split("/")).expanduser()
        if lab_info_path.exists():
            try:
                with lab_info_path.open() as fp:
                    lab_info = json.load(fp)
                lab_url = lab_info["lab_url"]
                has_proxy = lab_info["has_proxy"]
            except (OSError, KeyError):
                LOG.warning(f"Failed loading {lab_info_path}")
                pass

    if lab_url and has_proxy:
        reverse_prefix = f"/proxy/{port}"
        if lab_url.endswith("/"):
            lab_url = lab_url[:-1]
        return lab_url + reverse_prefix, reverse_prefix

    return f"http://localhost:{port}", ""


def _find_port(start: int = 8000, end: Optional[int] = None) -> int:
    """Find a port not in use in range *start* to *end*"""
    end = end if isinstance(end, int) and end >= start else start + 12000
    for port in range(start, end + 1):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex(("localhost", port)) != 0:
                return port
    raise RuntimeError("No available port found")
